/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LCURSOR_H_
#define _LCURSOR_H_

#include <lcl.h>

#include <lcl/llinkedcont.h>

class LBaseCollection;

/*!
 * \class LCursor
 * \brief A cursor class for collections
 * \author Jens Reimann
 * \ingroup Collections
 *
 * The cursor class may be used to iterate manually stepping through a list (LHeap, LSet, LMap).
 * A cursor points to an element in the list. You may step forward or backward in the list.
 *
 * Depending on the list you might also search for an item (by element, by key).
 *
 * Since the cursor only points to an element in the list it will be invalidated when the content of the
 * list is changed (remove, add, etc.). A call the isValid will return false then.
 *
 * LCursor is only a base class for all cursors. Cursors are implemented on a per-list basis since each list
 * might organise its data differently. Each cursor belongs to a specified list and may not be reassigned to
 * anther list. If you want your cursor to use another list, you need to delete it and create a new one!
 *
 * A cursor will may be created using the createCursor methods of LSet, LHeap, LMap
 */

/*!
 * \example cursor.cc
 * This example shows you how to use LCursor
 */

class LCL_EXPORT LCursor {
  friend class LBaseCollection;

 private:
  LCursor ( LCursor const & p1 ) {};

 protected:
  LBaseCollection * collection;

 public:

  bool valid;  

  LCursor ( LBaseCollection * coll );

  virtual ~LCursor ();

  virtual void invalidate ();              //!< Invalidates the cursor.
  virtual void obsolete ();

  virtual void setToNext () = 0;           //!< Sets the cursor to the next element
  virtual void setToPrev () = 0;           //!< Sets the cursor to the previous element
  virtual void setToFirst () = 0;          //!< Sets the cursor to the first element
  virtual void setToLast () = 0;           //!< Sets the cursor to the last element

  virtual bool isValid () const = 0;       //!< Checks if the cursor is valid

  virtual bool isFor ( const LBaseCollection * ) const;  //!< Checks if the cursor belongs to a list

  virtual LCursor * clone () = 0;
};

#include <lcl/lcollection.h>

/*!
 * \class LLinkedCursor
 * \ingroup Collections
 *
 * This is a cursor implementation for lists which use LLinkedSequence as data storage.
 *
 * For more information about cursors see LCursor. 
 */

template <class Type>
class LCL_EXPORT LLinkedCursor : public LCursor {
  friend class LBaseCollection;
 protected:

 public:

  LLinkedContainer<Type> * container;

  LLinkedCursor ( LLinkedCursor const & p1 );
  LLinkedCursor ( LBaseCollection * );

  virtual ~LLinkedCursor () {};

  virtual bool isValid () const;
  virtual void setToFirst ();
  virtual void setToNext ();
  virtual void setToLast ();
  virtual void setToPrev ();

  virtual LCursor * clone ();
};

template <class Type>
LLinkedCursor<Type>::LLinkedCursor ( LBaseCollection * _coll ) :
  LCursor ( _coll ),
  container ( NULL )
{  
}

template <class Type>
LLinkedCursor<Type>::LLinkedCursor ( LLinkedCursor const & p1 ) :
  LCursor ( p1.collection ),
  container ( p1.container )
{
}

template <class Type>
bool LLinkedCursor<Type>::isValid () const
{
  return (container != NULL) && (collection != NULL) && valid;
}

template <class Type>
void LLinkedCursor<Type>::setToFirst ()
{
  collection->findFirst ( this );
  valid = ( container != NULL );
}

template <class Type>
void LLinkedCursor<Type>::setToNext ()
{
  collection->findNext ( this );
  valid = ( container != NULL );
}

template <class Type>
void LLinkedCursor<Type>::setToPrev ()
{
  collection->findPrev ( this );
  valid = ( container != NULL );
}

template <class Type>
void LLinkedCursor<Type>::setToLast ()
{
  collection->findLast ( this );
  valid = ( container != NULL );
}

template <class Type>
LCursor * LLinkedCursor<Type>::clone ()
{
  return new LLinkedCursor ( *this );
}

// simple macro to do for-cursor loops easier
/*!
 * \param c Is either a LCursor* or LManagedPointer<Cursor>
 */
#define forcursor(c) for ( (c)->setToFirst(); (c)->isValid(); (c)->setToNext() )

#endif


