#ifndef _LFILEDESC_H_
#define _LFILEDESC_H_

#include <lcl.h>
#include <lcl/lvcs.h>

class LFileDesc;

#include <lcl/lselectloop.h>
#include <lcl/lemit.h>

/*!
 * \defgroup FileDescriptors File Descriptors
 */

/*!
 * \class LFileDesc
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A base class for all file descriptors
 * \ingroup FileDescriptors
 *
 * LFileDesc is a base class for all classes based on file descriptors. To know how to use this
 * class you must now some things about UNIX style file handling:
 *
 * Nearly everything in a UNIX environment is based on file descriptors. They are like handles just
 * a bit more generic. When you open a file you will receive a file descriptor (file handle) which
 * you may use to access files. But that's only one small part of file descriptors.
 *
 * File descriptors are a bit object orientated. There are two types character based and block based
 * file descriptors. Characeter based file descriptors allow you to access a (let's simply call it)
 * stream characterwise. You have access to each character in the stream.
 *
 * Block based file descriptors only allow to you access you stream in blocks, which means that you
 * cannot access each character of the "stream".
 *
 * Character based file descriptors are mostly files, sockets but they may also be serial ports,
 * pipes, etc. Block based file descriptors are e.g. tape drives, hard discs, SCSI devices. These
 * may only be accessed by sending/reading a block of data to it/from it.
 *
 * Nearly all file descriptors are created by using the ::open call. One parameter of the ::open call
 * is the filename. But a filename may also be a name of a device (e.g. /dev/ttyS0 for the first
 * serial port). You see, it's easy to open the serial port under UNIX ;-)
 *
 * As already mentioned file descriptors are OO based. Most of them support the ::read and ::write
 * calls. Many ::open. All ::close.
 *
 * You can for example close any file descriptor (no difference between hard disc, serial port or 
 * file) using ::close. That's easy, isn't it? *gg* So you do not need to differ!
 *
 * On most operating systems file descriptors are simply implemented using int as identifier. All
 * file descriptors application wide share the same ID pool.
 *
 * There are some "default" file descriptors 0, 1 and 2. Which are (nearly) always available for
 * an application. They are mapped to cin, cout and cerr (or stdin, stdout and stderr). But, as you
 * already know, these need not to be files, they may be pipes, the console or anything else.
 * (btw. this is how your shell does redirection ;-) )
 *
 * You might ask, why not use a different type of handle for each type of "thing" you access.
 * Using the same handle for all function calls might result in a wrong file descriptor
 * be passed to a function which does support the required operation (e.g. passing a file
 * handle to a ::connect call which should connect a socket). Ok, just have a look behind the scenes:
 *
 * Each file descriptor supports something open, close, read and write. That's it (nearly *g*)
 * Special operations like connect, seek, etc. are realized using sysctl (or an equivalent) which
 * will receive the file descriptor and the requested operation. The corresponding 
 * operating system module will then check if the requested operation is supported by this
 * type of file descriptor, if not you will get an error! Very simple ;-)
 *
 * Now, back to LFileDesc (where we started). LFileDesc is a base class for all these different
 * file descriptors, whethere they are files, sockets or what so ever...
 *
 * In most cases you will not use LFileDesc but for example LFile to access a file or LSocket to
 * access a socket.
 *
 * One of the benfits using this concept is, that all these file descriptors may use the ::select 
 * (or ::poll) function. Which waits for an event occuring at the file descriptor (e.g. data
 * available for read ...). This is also one of the major features of the LFileDesc class. All
 * instances of LFileDesc (or derieved from LFileDesc) may use the LSelectLoop class. Once
 * registered with a LSelectLoop instance you may call LSelectLoop::loop and wait for an event
 * occuring at any file descriptor. You will then be notified using (e.g. LEmitter0). So you
 * simply set up your file descritors, register them with LSelectLoop and wait for an event.
 * For more information about event handling using file descriptors see LSelectLoop!
 *
 * The whole file descriptor thing is abstraced using the LFDManager class to support this
 * behaviour under WIN32 although (up to now) it is not fully supported)
 *
 * The lifetime of a file descritor is the same as the lifetime of the assigned object instance!
 */

class LCL_EXPORT LFileDesc {
  LVCS_DECLARE;
  friend class LSelectLoop;

 protected:
  LFileDescriptor fd;                 //!< The operating system file descriptor
  /*  unsigned long eMask;*/                //!< The effective event mask
  bool auto_close;

  void copyFrom ( LFileDesc const & );
  void assign ( LFileDescriptor );
  virtual void emitReadEvent ();
  virtual void emitWriteEvent ();
  virtual void emitXcptEvent ();

 public:

  void reg ( LSelectLoop * );         //!< Register the file descriptor with an LSelectLoop
  void unreg ( LSelectLoop * );       //!< Unregister the file descriptor with an LSelectLoop

  LEmitter1<LFileDesc*> readEvent;    //!< The event which will be fired when a read event occurs
  LEmitter1<LFileDesc*> writeEvent;    //!< The event which will be fired when a write event occurs
  LEmitter1<LFileDesc*> xcptEvent;    //!< The event which will be fired when a exception event occurs
  
  LFileDescriptor getHandle () const;
  enum EventMask { eventRead = 0x01,  //!< The event mask value for read events 
		   eventWrite = 0x02, //!< The event mask value for write events 
		   eventXcpt = 0x04   //!< The event mask value for exception events 
  };

  LFileDesc ();                        //!< Create a new invalid file descriptor
  LFileDesc ( LFileDescriptor );       //!< Create a new file descriptor from an external already existing one
  LFileDesc ( LFileDesc const & );     //!< Copy a file descriptor from an existing one

  virtual ~LFileDesc ();

  void close ();                       //!< Close the file descriptor

  bool isValid () const;               //!< Check if the file descriptor is valid

  void setEventMask ( unsigned long ); //!< Set the event mask for a file descriptor (see LSelectLoop)
  unsigned long eventMask () const;    //!< Retrieve the event mask

  int write ( void const *, size_t len ); //!< Write data to the file descriptor
  int read ( void *, size_t len );        //!< Read data from the file descriptor

  int write ( LBuffer const & );          //!< Write data to the file descriptor
  int read ( LBuffer &, size_t len );     //!< Read data from the file descriptor

  void setAutoClose ( bool _flag );       //!< Disable the automatic destruction of the file desctiptor

  LFileDesc & operator = ( LFileDesc const & p1 );

#ifndef _WIN32
  void setBlocking ( bool state = true );
  bool getBlocking () const;
#endif

  /*
  int fcntl ( int cmd ) const;
  int fcntl ( int cmd, long arg );
  */
};

#endif
