/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LINTERPRETER_H_
#define _LINTERPRETER_H_

#include <lcl/lhash.h>
#include <lcl/lpointer.h>
#include <lcl/ltable.h>
#include <lcl/lstring.h>
#include <lcl/lheap.h>
#include <lcl/lvcs.h>
#include <lcl/lvariant.h>
#include <lcl/lstack.h>

/*!
 * \defgroup Scripting Scripting
 */

/*!
 * \class LInterpreter
 * \brief A interpreter class for executing parsed scripts
 * \author Jens Reimann <ctron@dentrassi.de>
 * \todo Could have more real exceptions
 * \ingroup Scripting
 *
 * A master class for controlling and executing parsed scripts. The scripts must
 * first be parsed by a parser derieved from LParser.
 *
 * A parser will return a Sequence which contains the execution path of a script.
 *
 * To create a new command derieve a class from Command and register it using
 * LInterpreter::registerCommand
 */

class LCL_EXPORT LInterpreter {
  LVCS_DECLARE;

public:
  class LCL_EXPORT Command;
  class LCL_EXPORT Sequence;

protected:
  LHash<LString, LManagedPointer<Command> > commandHash;   //!< Holds the referenced to registerd commands
  //  LHash<LString, LHash<LString,LString> > globalContextHash;
public:

  class LCL_EXPORT Exception : public LBasicException {
  public:
    enum Reason { UNKOWN, CMD_INVALID, PARAM_CNT };
  protected:    
    Reason reason;
    LString tmp;
  public:
    Exception ( Reason _reason, LString const & _tmp );
    void log ( ) const;
    virtual ~Exception();

    LString getArg1 () const;
  };

  class LCL_EXPORT Context {
  protected:
    LInterpreter & inter;
    LHash<LString, LHash<LString,LVariant> > globalHash;
    LStack < LHash < LString, LHash < LString, LVariant > > > localHash;
  public:

    Context ( LInterpreter & _inter );
    virtual ~Context ();

    void pushStack ();
    void popStack ();

    LHash<LString,LVariant> & getLocal ( LString const & name );
    LHash<LString,LVariant> & getGlobal ( LString const & name );

    void clear ();

    LVariant run ( LManagedPointer<Sequence> sequence );

    LInterpreter & getInterpreter ();
  };  

  /*!
   * \class Command
   * \brief Virtual base class for a command
   * \ingroup Scripting
   */
  class LCL_EXPORT Command {
  protected:
    LString name;
  public:
    Command ( LString const & _name );
    virtual ~Command ();
    virtual LVariant exec ( Context & context, LTable<LVariant> const & args ) = 0;

    LString getName () const;

    virtual bool isValid () const;
  };

  class LCL_EXPORT ScriptCommand : public Command {
  protected:
    LManagedPointer<Sequence> sequence;

  public:
    ScriptCommand ( LString const & _name, LManagedPointer<Sequence> _seq );
    virtual ~ScriptCommand ();
    virtual LVariant exec ( Context & contex, LTable<LVariant> const & args );

    virtual bool isValid () const;
  };

  /*!
   * \class Sequence
   * \brief Virtual base class for a sequence
   * \ingroup Scripting
   */
  class LCL_EXPORT Sequence {
  protected:
    LHeap < LManagedPointer < Sequence > > args;
  public:
    virtual ~Sequence ();
    virtual void addParameter ( LManagedPointer<Sequence> _arg );
    virtual LVariant exec ( Context & ) = 0;
  };

  class LCL_EXPORT Sequence_Command : public Sequence {
  protected:
    LString name;
  public:
    Sequence_Command ( LString const & _name );
    virtual ~Sequence_Command ();
    virtual LVariant exec ( Context & );
  };

  class LCL_EXPORT Sequence_Const : public Sequence {
  protected:
    LVariant value;
  public:
    Sequence_Const ( LVariant const & _value );
    virtual ~Sequence_Const ();
    virtual LVariant exec ( Context & );
  };

  class LCL_EXPORT Sequence_If : public Sequence {
  protected:
  public:
    virtual ~Sequence_If ();
    virtual LVariant exec ( Context & );
  };

  class LCL_EXPORT Sequence_For : public Sequence {
  protected:
  public:
    virtual ~Sequence_For ();
    virtual LVariant exec ( Context & );    
  };

  class LCL_EXPORT Sequence_While : public Sequence {
  protected:
  public:
    virtual ~Sequence_While ();
    virtual LVariant exec ( Context & );    
  };

  class LCL_EXPORT Sequence_Group : public Sequence {
  protected:
  public:
    virtual ~Sequence_Group ();
    virtual LVariant exec ( Context & );
  };

  virtual ~LInterpreter ();

  void registerCommand ( LManagedPointer<Command> command );   //!< Register a new command
  void unregisterCommand ( LString const & name );             //!< Unregister a new command
  LManagedPointer<Command> getCommand ( LString const & name ) const; //!< fetch a command
  LHash<LString,LVariant> & getContextHash ( LString const & name ); //!< fetch the context hash for this interpreter

  LVariant run ( LManagedPointer<Sequence> sequence );   //!< run the parsed sequence

  //  void clear (); //!< clear all running context and states
};



#define _BASE_COMMAND_DEFINE(s) class cmd_##s : public LInterpreter::Command { \
       protected:\
       public:\
         cmd_##s (); \
         virtual ~cmd_##s () {} ; \
         virtual LVariant exec ( LInterpreter::Context & context, LTable<LVariant> const & args ); \
       };

#define _BASE_COMMAND_DEFINE_RES(s) class reserved_##s : public LInterpreter::Command { \
       protected:\
       public:\
         reserved_##s (); \
         virtual ~reserved_##s () {} ; \
         virtual LVariant exec ( LInterpreter::Context & context, LTable<LVariant> const & args ); \
       };


#define _BASE_COMMAND_CONSTRUCTOR(s) cmd_##s::cmd_##s () : LInterpreter::Command ( #s ) {}
#define _BASE_COMMAND_CONSTRUCTOR_ALT(s,a) cmd_##s::cmd_##s () : LInterpreter::Command ( a ) {}
#define _BASE_COMMAND_BODY_START(s) LVariant cmd_##s::exec ( LInterpreter::Context & context, LTable<LVariant> const & args ) { LTRACE ( "LCommands", "cmd_" #s "::exec ( ctx, args [" + LString(args.getSize() ) + "] )"  );
#define _BASE_COMMAND_BODY_END(s) }
#define _BASE_COMMAND_REGISTER(s) inter.registerCommand ( new cmd_##s );
#define _BASE_COMMAND_REGISTER_EXP(s) inter.registerCommand ( new s );

#define _BASE_COMMAND_CONSTRUCTOR_RES(s) reserved_##s::reserved_##s () : LInterpreter::Command ( #s ) {}
#define _BASE_COMMAND_BODY_START_RES(s) LVariant reserved_##s::exec ( LInterpreter::Context & context, LTable<LVariant> const & args )
#define _BASE_COMMAND_BODY_END_RES(s)
#define _BASE_COMMAND_REGISTER_RES(s) inter.registerCommand ( new reserved_##s );

#define _BASE_COMMAND_SIMPLE_1(s, c) _BASE_COMMAND_CONSTRUCTOR(s) _BASE_COMMAND_BODY_START(s) { return (c); } _BASE_COMMAND_BODY_END(s);
#define _BASE_COMMAND_SIMPLE_1_RES(s, c) _BASE_COMMAND_CONSTRUCTOR_RES(s) _BASE_COMMAND_BODY_START_RES(s) { return (c); } _BASE_COMMAND_BODY_END_RES(s);

#define _BASE_COMMAND_SIMPLE_2(s, c) _BASE_COMMAND_CONSTRUCTOR(s) _BASE_COMMAND_BODY_START(s) { (c); return ""; } _BASE_COMMAND_BODY_END(s);
#define _BASE_COMMAND_SIMPLE_2_RES(s, c) _BASE_COMMAND_CONSTRUCTOR_RES(s) _BASE_COMMAND_BODY_START_RES(s) { (c); return ""; } _BASE_COMMAND_BODY_END_RES(s);

#define _BASE_COMMAND_SIMPLE_3(s, c) _BASE_COMMAND_CONSTRUCTOR(s) _BASE_COMMAND_BODY_START(s) { c; } _BASE_COMMAND_BODY_END(s);
#define _BASE_COMMAND_SIMPLE_3_RES(s, c) _BASE_COMMAND_CONSTRUCTOR_RES(s) _BASE_COMMAND_BODY_START_RES(s) { c; } _BASE_COMMAND_BODY_END_RES(s);


#endif
