/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LLINKEDSEQUENCE_H_
#define _LLINKEDSEQUENCE_H_

//#include <assert.h>

#include <lcl.h>
#include <lcl/lxcpt.h>

/*!
 * \class LLinkedSequence
 * \ingroup Collections
 * \author Jens Reimann <ctron@dentrassi.de>b
 *
 * This is a very generic implementation of a double linked list. It is mostly used by the
 * collections (LHeap, LSet, LMap, LHash). A linked sequence only knows about containers and not
 * about the actual data which should be stored in a sequence. It is designed as a helper class for
 * high level collections.
 *
 * If you want to store data in collections use LHeap, LSet, LMap or LHash instead!
 *
 * The linked list stores data items in containers. The container has to be specified when instantiating
 * the class. In most cases LLinkedContainer will be used.
 *
 * A container needs a next and prev pointer which will point to the next or previous item in the linked list.
 * If either next or prev is NULL the container is head (or tail) of the sequence.
 *
 * It also will define (otherwise it would be very stupid) a data member (which will not be used by
 * LLinkedSequence) which will reflect the data wrapped in the container.
 *
 */

template <class Type, class Container>
class LCL_EXPORT LLinkedSequence {
 public:
  Container *head, *tail;
  unsigned long items;

 public:
  LLinkedSequence ();
  LLinkedSequence ( LLinkedSequence<Type,Container> const & );
  virtual ~LLinkedSequence ();

  virtual void addContainer ( Container * );              //!< Add a new container to the sequence
  virtual void removeContainer ( Container * );           //!< Remove a container which is in the sequence
  virtual void removeContainerAndDelete ( Container * );  //!< Remove a container which is in the sequence and delete it afterwards

  //! The assignment operator of the sequence
  LLinkedSequence<Type, Container> const & operator = ( LLinkedSequence<Type, Container> const & );

  //! Clears the sequence and frees all containers
  virtual void clear ();
};

template <class Type, class Container>
LLinkedSequence<Type, Container>::LLinkedSequence () :
  head ( NULL ), tail ( NULL ), items ( 0L )
{
}

template <class Type, class Container>
LLinkedSequence<Type,Container>::LLinkedSequence ( LLinkedSequence<Type,Container> const & p1 ) :
  head ( NULL ), tail ( NULL ), items ( 0L )
{
  Container * current = p1.head;
  while ( current ) {
    addContainer ( new Container ( *current ) );
    current = current->next;
  }
}

template <class Type, class Container>
LLinkedSequence<Type, Container>::~LLinkedSequence ()
{
  clear ();
}

template <class Type, class Container>
LLinkedSequence<Type, Container> const & LLinkedSequence<Type,Container>::operator = ( LLinkedSequence<Type, Container> const & p1 )
{
  clear ();

  Container * current = p1.head;
  while ( current ) {
    addContainer ( new Container ( *current ) );
    current = current->next;
  }

  return *this;
}

/*!
 * \param container The container which should be appended
 *
 * It will append a new container to the sequence. The container will then be maintained by the sequence.
 * The object may not be deleted while it is in the sequence!
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::addContainer ( Container * container )
{
  if ( container == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );
  
  if ( head == NULL ) {
    head = tail = container;
  } else {

    lcl_assert ( tail != NULL );

    container->prev = tail;
    container->next = NULL;
    tail->next = container;    
    tail = container;

  }

  items++;
}

/*!
 * All entries which a currently in the sequence will be removed. All containers will be deleted and the
 * memory will be freed.
 *
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::clear ( )
{
  Container *current = head, *temp;
  while ( current ) {
    temp = current->next;
    delete current;
    current = temp;
  }

  head = tail = NULL;
  items = 0;
}


// ***********************************************************************************************
// **  Delete container from linked list
// ***********************************************************************************************

/*!
 * \param container The container which should be removed from the sequence
 *
 * The specified container will be removed from the sequence. The container will not be deleted and
 * the memory is not freed! You have do to this manually using delete!
 *
 * If you want to free the container automatically you need to call LLinkedSequence::removeContainerAndDelete
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::removeContainer ( Container * container )
{
  Container * current = head;

  while ( current ) {
    if ( current == container ) {

      if ( current->prev ) {
	current->prev->next = current->next;
      } else {
	head = current->next;
      }
      if ( current->next ) {
	current->next->prev = current->prev;       
      } else {
	tail = current->prev;
      }
      
    }
    current = current->next;
  }

  items--;
}

/*!
 * \param container The container which should be removed from the sequence
 *
 * The specified container will be removed from the sequence. The container will be deleted and the memory
 * freed.
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::removeContainerAndDelete ( Container * container )
{
  removeContainer ( container );
  delete container;
}

#endif

