#ifndef _LMAP_H_
#define _LMAP_H_

#include <lcl.h>

#include <lcl/lcollection.h>
#include <lcl/llinkedcont.h>
#include <lcl/llinkedseq.h>

#include <lcl/literator.h>
#include <lcl/lcursor.h>

/**
 * \class LMap
 * \brief A collection class which contains elements identified by a key element
 * \author Jens Reimann <ctron@dentrassi.de>
 * \version 0.2
 * \date 1999
 * \bug Should be very bug free.
 * \ingroup Collections
 *
 * This is a collection with unsorted elements identified by a key element. The key element must be unique in this collection.
 * To retrieve the key element the collection will call a \e key \e function defined as \c Element \c key \c ( \c Type \c const \c & \c ) which has to be defined for the class stored in the collection.
 * 
 * It is possible to retreive elements by their key.
 *
 * A LLinkedSequence is use as storage.
 */

template <class Type, class Key>
class LCL_EXPORT LMap : public LCollection<Type> {
 protected:
  LLinkedSequence<Type, LLinkedContainer<Type> > sequence;   //!< The storage for this class

  LLinkedCursor<Type> * baseCursor ( LCursor * ) const;      //!< Converts the given base cursor to a cursor of this class

 public:
  LMap ();                                                   //!< Constructor
  LMap ( LMap<Type,Key> const & );                           //!< Copy Constructor
  virtual ~LMap();                                           //!< Destructor

  virtual LCursor * createCursor () const;                   //!< Creates a new cursor which is owned by this collection

  virtual bool add ( Type const & );                         //!< Add a new element to this collection
  virtual unsigned long entries () const;                    //!< Retrieves the number of elements in this collection

  virtual void findFirst ( LCursor * ) const;                //!< Sets the cursor to the first element
  virtual void findNext ( LCursor * ) const;                 //!< Sets the cursor to the next element
  virtual void findPrev ( LCursor * ) const;                 //!< Sets the cursor to the previous element
  virtual void findLast ( LCursor * ) const;                 //!< Sets the cursor to the last element

  virtual Type & elementAt ( LCursor * );                    //!< Retrieves the element at this cursor position
  virtual Type const & elementAt ( LCursor * ) const;        //!< Retrieves the element as const at this cursor position
  virtual Type & elementWithKey ( Key const & );             //!< Retrieves the element with the given key
  virtual Key keyAt ( LCursor * ) const;                     //!< Retrieves the key at this cursor position

  virtual void locateElementWithKey ( Key const &, LCursor * ) const; //!< Sets the cursor to the element with the given key
  virtual LLinkedContainer<Type> * findContainerWithKey ( Key const & ) const;

  virtual bool containsElementWithKey ( Key const & ) const; //!< Checks if the collection contains an element with the given key
  virtual bool replaceElementWithKey ( Key const &, Type const & ); //!< Replaces the element with a given key

  virtual bool removeElementAt ( LCursor * );                //!< removes the element at cursor position from the collection
  virtual bool removeElementWithKey ( Key const & );         //!< removes the element with the given key from the collection

  virtual void allElementsDo ( LIterator<Type> & );          //!< Iterate through all elements

  virtual void clear ();                                     //!< Clear the collection

  virtual LMap<Type,Key> const & operator = ( LMap<Type,Key> const & ); //!< Assign operator

  void dump () const;
};

template <class Type, class Key>
LMap<Type,Key>::LMap ()
{  
}

template <class Type, class Key>
LMap<Type,Key>::LMap ( LMap<Type,Key> const & p1 ) :
  sequence ( p1.sequence )
{  
  //  cout << "COPY\n";
}

template <class Type, class Key>
LMap<Type, Key>::~LMap()
{
  clear();
}

template <class Type, class Key>
LMap<Type,Key> const & LMap<Type,Key>::operator = ( LMap<Type,Key> const & p1 )
{
  clear();
  sequence = p1.sequence;
  return *this;
}

template <class Type, class Key>
bool LMap<Type,Key>::add ( Type const & item )
{
  LLinkedContainer<Type> * container = new LLinkedContainer<Type>( item );
  sequence.addContainer ( container );
  invalidate ();
  return true;
}

template <class Type, class Key>
unsigned long LMap<Type,Key>::entries () const
{
  return sequence.items;
}

template <class Type, class Key>
LLinkedCursor<Type> * LMap<Type,Key>::baseCursor ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) )
    return (LLinkedCursor<Type>*)cursor;
  return NULL;
}

template <class Type, class Key>
void LMap<Type,Key>::dump ( ) const
{
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    NAMESPACE_STD(cout) << current->item << "\n";
    current = current->next;
  }
}

template <class Type, class Key>
void LMap<Type,Key>::findFirst ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.head;
  }
}

template <class Type, class Key>
void LMap<Type,Key>::findNext ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->next;
  }
}

template <class Type, class Key>
void LMap<Type,Key>::findLast ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.tail;
  }
}

template <class Type, class Key>
void LMap<Type,Key>::findPrev ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->prev;
  }
}

template <class Type, class Key>
LCursor * LMap<Type,Key>::createCursor () const
{
  return new LLinkedCursor<Type>( (LBaseCollection*)this);
}

template <class Type, class Key>
Type & LMap<Type,Key>::elementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1; 
  MSTHROW(Type);
}

template <class Type, class Key>
Type const & LMap<Type,Key>::elementAt ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1;
  MSTHROW(Type);
}

template <class Type, class Key>
Key LMap<Type,Key>::keyAt ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return key ( baseCursor ( cursor )->container->item );
  }
  throw 1;
  MSTHROW(Key);
}

template <class Type, class Key>
void LMap<Type,Key>::locateElementWithKey ( Key const & findKey, LCursor * cursor ) const
{
  if ( !cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = NULL;
  } else {

    LLinkedContainer<Type> * current = sequence.head;

    while ( current ) {
      if ( key ( current->item ) == findKey ) {
	baseCursor(cursor)->container = current;
	baseCursor(cursor)->valid = true;
	return;
      }
      current = current->next;
    }

    baseCursor(cursor)->container = NULL;

  }
}

template <class Type, class Key>
LLinkedContainer<Type> * LMap<Type,Key>::findContainerWithKey ( Key const & findKey) const
{
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    if ( key ( current->item ) == findKey ) {
      return current;
    }
    current = current->next;
  }

  return NULL;
}

template <class Type, class Key>
inline Type & LMap<Type,Key>::elementWithKey ( Key const & findKey )
{
  LLinkedContainer<Type> * container = findContainerWithKey ( findKey );
  if ( container ) {
    return container->item;
  } else {
    throw 1;
	MSTHROW(Type);
  }
}

template <class Type, class Key>
bool LMap<Type,Key>::replaceElementWithKey ( Key const & findKey, Type const & item )
{
  LLinkedContainer<Type> * container = findContainerWithKey ( findKey );
  if ( container ) {
    container->item = item;
    invalidate();
    return true;
  } else {
    return false;
  }
}

template <class Type, class Key>
bool LMap<Type,Key>::containsElementWithKey ( Key const & findKey ) const
{
  return findContainerWithKey ( findKey ) != NULL;
}

template <class Type, class Key>
bool LMap<Type,Key>::removeElementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    sequence.removeContainer ( baseCursor(cursor)->container );
    delete baseCursor(cursor)->container;
    invalidate ();
    return true;
  } else {
    return false;
  }
}

template <class Type, class Key>
bool LMap<Type,Key>::removeElementWithKey ( Key const & findKey )
{
  LLinkedContainer<Type> * container = findContainerWithKey ( findKey );
  if ( container ) {
    sequence.removeContainer ( container );
    delete container;
    invalidate();
    return true;
  } else {
    return false;
  }
}

template <class Type, class Key>
void LMap<Type,Key>::clear () {
  sequence.clear ();
}

template <class Type, class Key>
void LMap<Type, Key>::allElementsDo ( LIterator<Type> & p_iterator )
{
  LLinkedContainer<Type> * current = sequence.head;
  while ( current ) {
    p_iterator.applyTo ( current->item );
    current = current->next;
  }
}


#endif

