/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LSET_H_
#define _LSET_H_

#include <lcl.h>

#include <lcl/lcollection.h>
#include <lcl/llinkedcont.h>
#include <lcl/llinkedseq.h>
#include <lcl/literator.h>
#include <lcl/lcursor.h>
#include <lcl/lpointer.h>

/**
 * \class LSet
 * \brief A collection class which contains unique, unsorted elements.
 * \author Jens Reimann <ctron@dentrassi.de>
 * \ingroup Collections
 *
 * This is a unsorted collection which contains unique elements. The elements used with this collection need to
 * have the operator == defined. It is possible to retreive an object directly from the collection by comparsion. 
 *
 * A LLinkedSequence is use as storage.
 */


/*!
 * \example lcoll_sample_1.cc
 * This example gives you a clue about how to use LCL Collections
 */


template <class Type>
class LCL_EXPORT LSet : public LCollection<Type> {
 protected:
  LLinkedSequence<Type, LLinkedContainer<Type> > sequence;     //!< The storage for this class

  LLinkedCursor<Type> * baseCursor ( LCursor * ) const;        //!< Converts the given base cursor to a cursor of this class

 public:
  LSet ();                                                     //!< Constructor
  LSet ( LSet<Type> const & );                                 //!< Copy Constructor
  virtual ~LSet();                                             //!< Destructor

  virtual LCursor * createCursor () const;                     //!< Creates a new cursor which is owned by this collection

  virtual bool add ( Type const & );                           //!< Add a new element to this collection
  virtual unsigned long addAllFrom ( LSet<Type> const & );     //!< Adds all elements from the given collection
  virtual unsigned long entries () const;                      //!< Retrieves the number of elements in this collection

  virtual void findFirst ( LCursor * ) const;                  //!< Sets the cursor to the first element
  virtual void findNext ( LCursor * ) const;                   //!< Sets the cursor to the next element
  virtual void findPrev ( LCursor * ) const;                   //!< Sets the cursor to the previous element
  virtual void findLast ( LCursor * ) const;                   //!< Sets the cursor to the last element

  virtual Type & elementAt ( LCursor * );                      //!< Retrieves the element at this cursor position
  virtual Type const & elementAt ( LCursor * ) const;          //!< Retrieves the element as const at this cursor position

  virtual bool containsElement ( Type const & item ) const;    //!< Checks if the collection contains the given element

  virtual bool locateElement ( Type const & element, LCursor * ); //!< Sets the cursor to the position of the given element

  virtual bool removeElementAt ( LCursor * );               //!< Removes the element at the cursor position from the collection
  virtual bool removeElement ( Type const & );              //!< Remove the given element from the collection

  virtual void allElementsDo ( LIterator<Type> & );            //!< Iterates through all elements

  virtual void clear ();                                       //!< Clear the collection

  void operator += ( LSet<Type> const & );                     //!< Adds all elements from the given collection

  void dump () const;
};

template < class Type >
/*LCL_EXPORT*/ bool equal ( Type const & p1, Type const & p2 );

template < class Type >
/*LCL_EXPORT*/ bool equal ( Type const & p1, Type const & p2 )
{
  return p1 == p2;
}

template <class Type>
LSet<Type>::LSet ()
{  
}

template <class Type>
LSet<Type>::LSet ( LSet const & p1 ) :
  sequence ( p1.sequence )
{
}

template <class Type>
LSet<Type>::~LSet()
{
}

template <class Type>
bool LSet<Type>::add ( Type const & item )
{
  LLinkedContainer<Type> * container = new LLinkedContainer<Type>( item );
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    if ( equal ( current->item, item ) )
      return false;
    current = current->next;
  }

  sequence.addContainer ( container );
  invalidate ();
  return true;
}

template <class Type>
inline unsigned long LSet<Type>::addAllFrom ( LSet<Type> const & p1 )
{
  int cnt = 0;
  LLinkedContainer<Type> * current = p1.sequence.head;
  
  while ( current ) {    
    add ( current->item );
    cnt ++;
    current = current->next;
  }

  return cnt;
}

template <class Type>
unsigned long LSet<Type>::entries () const
{
  return sequence.items;
}

template <class Type>
void LSet<Type>::clear ()
{
  sequence.clear ();
  invalidate ();
}

template <class Type>
LLinkedCursor<Type> * LSet<Type>::baseCursor ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) )
    return (LLinkedCursor<Type>*)cursor;
  return NULL;
}

template <class Type>
void LSet<Type>::dump ( ) const
{
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    //    cout << current->item << "\n";
    current = current->next;
  }
}

template <class Type>
void LSet<Type>::findFirst ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.head;
    baseCursor(cursor)->valid = true;
  }
}

template <class Type>
void LSet<Type>::findNext ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->next;
  }
}

template <class Type>
void LSet<Type>::findLast ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.tail;
    baseCursor(cursor)->valid = true;
  }
}

template <class Type>
void LSet<Type>::findPrev ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->prev;
  }
}

template <class Type>
LCursor * LSet<Type>::createCursor () const
{
  return new LLinkedCursor<Type>( (LBaseCollection*)this);
}

template <class Type>
Type & LSet<Type>::elementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1;
  MSTHROW(Type);
}

template <class Type>
Type const & LSet<Type>::elementAt ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1;
  MSTHROW(Type);
}

template <class Type>
inline bool LSet<Type>::locateElement ( Type const & item, LCursor * cursor )
{
  if ( cursor->isFor ( this ) ) {
    LLinkedContainer<Type> * current = sequence.head;
    while ( current ) {
      if ( equal ( current->item, item ) ) {
	baseCursor ( cursor )->container = current;
	baseCursor ( cursor )->valid = true;
	return true;
      }
      current = current->next;
    }
    baseCursor ( cursor )->valid = false;
    return false;
  }

  throw 1;
  MSTHROW(bool);
}

template <class Type>
inline bool LSet<Type>::containsElement ( Type const & item ) const
{
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    if ( equal ( current->item, item ) )
      return true;
    current = current->next;
  }
  return false;
}

template <class Type>
inline bool LSet<Type>::removeElementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    sequence.removeContainer ( baseCursor(cursor)->container );
    delete baseCursor(cursor)->container;
    invalidate ();
    return true;
  } else {
    return false;
  }
}

template <class Type>
inline bool LSet<Type>::removeElement ( Type const & item )
{
  LLinkedContainer<Type> * current = sequence.head;
  while ( current ) {
    if ( equal ( current->item, item ) ) {
      sequence.removeContainer ( current );
      delete current;
      invalidate ();
      return true;
    }
    current = current->next;
  }
  return false;
}

template <class Type>
void LSet<Type>::allElementsDo ( LIterator<Type> & p_iterator )
{
  LLinkedContainer<Type> * current = sequence.head;
  while ( current ) {
    p_iterator.applyTo ( current->item );
    current = current->next;
  }
}

template <class Type>
inline void LSet<Type>::operator += ( LSet<Type> const & p1 )
{
  addAllFrom ( p1 );
}


template < class Type, class Serializer >
Serializer & operator << ( Serializer & ser, LSet<Type> const & set );

template < class Type, class Serializer >
Serializer & operator << ( Serializer & ser, LSet<Type> const & set )
{
  Serializer _ser;
  LManagedPointer < LCursor > cursor = set.createCursor();
  forcursor ( cursor )
    {
      _ser << set.elementAt ( cursor );
    }
  ser << _ser;
  return ser;
}

template < class Type, class Serializer >
Serializer & operator >> ( Serializer & ser, LSet<Type> & set )
{

  set.clear ();

  Serializer _ser;
  ser >> _ser;

  try
    {
      while ( true )
	{
	  Type item;
	  _ser >> item;
	  set.add ( item );
	}
    }
  catch ( ... )
    {
    }
  return ser;
}



#endif
