/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LSTRING_H_
#define _LSTRING_H_

/*!
 * \defgroup String String
 */

/*!
 * \class LString
 * \brief Complex-Generic string class
 * \author Jens Reimann <ctron@dentrassi.de>
 * \ingroup String
 * \warning Using LString::word can be very slow! Use LStringTokenizer when you
 * heavily process words!
 * \warning Token handling is now moved to LStringTokenizer. It is still included in LString for compatibility reasons.
 * Also it is sometimes much faster than first creating a LStringTokenizer instance.
 *
 * \note Words start with index 0
 *
 * This class support serialzation using LSerialize
 */

#include <lcl.h>

class LCL_EXPORT LString;

//! A special hash function for LStrings
LCL_EXPORT unsigned long hash ( LString const & item, unsigned long hash_size );

#include <lcl/lstringbase.h>

#include <lcl/lvcs.h>
#include <lcl/lobject.h>
#include <lcl/lutils.h>

#include <errno.h>
#include <string>
#include <stdarg.h>

class LCL_EXPORT LRegEx;
// 
// Words start with #0
//

class LCL_EXPORT LBuffer;
template <class Key, class Element>
class LCL_EXPORT LHash;

#include <lcl/lstringbase.h>

class LCL_EXPORT LString  : public LStringBase {
  LVCS_DECLARE;
 private:

  /*
  void fitMem ( unsigned long top );

  static const int C_block_size;
  */

 protected:
  /*  unsigned long ptr_space;*/ //!< The number of bytes allocated
  unsigned long ptr_used;  //!< The number of bytes used
  /*  char * string_ptr;*/ //!< The pointer to our master string
  
  char * delims;
  char * qmarks;

  static char * defaultDelims;
  static char * defaultQMarks;

  virtual void clear ();

 public:

  class LCL_EXPORT Replacer;

  // Ctors
  // -----------

  /*!
   * \name Constructors & Destructor
   */
  //@{
  LString ();                               //!< Default constructor. Creates an empty NULL string
  LString ( const char );                   //!< Constructor. Creates a one character string
  LString ( const char * );                 //!< Constructor which creates a copy of a string
  LString ( LString const & );              //!< Copy constructor
  LString ( LBuffer const & );              //!< Creates a LString object from a LBuffer
  LString ( unsigned long p1 );             //!< Converts a unsigned long to a string and makes a LString of it
  LString ( signed long );                  //!< Converts a long to a string and makes a LString of it

  LString ( signed int );                   //!< Converts a int to a string and makes a LString of it
  LString ( unsigned int );                 //!< Converts a unsigned int to a string and makes a LString of it

  LString ( signed short int );             //!< Converts a short int to a string and makes a LString of it
  LString ( unsigned short int );           //!< Converts a unsigned short int to a string and makes a LString of it

  LString ( lint64 p1 );         //!< Converts a signed long long int to a string and makes a LString of it
  LString ( luint64 p1 );       //!< Converts a unsigned long long int to a string and makes a LString of it

  LString ( double );                       //!< Converts a double to a string and makes a LString of it. Default conversion is sprintf("%f")

  // Dtor
  // -----------

  virtual ~LString ();                      //!< Destructor. Destroys the LString object and frees the memory which was allocated by it.
  //@}

  // operators
  // -----------

  LString & operator = ( LString const & );
  LString & operator = ( const char * );
  LString & operator = ( char value );
  LString & operator = ( double value );
  LString & operator = ( long value );

  //  template < class any_string > LString & operator = ( any_string const & str1 ) { clear (); append ( (const char*)str1 ); return *this; };
  
  LString operator + ( LString & ) const;
  LString operator + ( LString const & ) const;
  LString operator + ( const char * ) const;
  LString operator + ( char * ) const;

  LString & operator += ( LString const & );
  LString & operator += ( const char * );
  LString & operator += ( char );

  bool operator == ( LString const & ) const;
  bool operator == ( const char * ) const;
  bool operator != ( LString const & ) const;
  bool operator != ( const char * ) const;

  bool operator ! ( ) const;

  char & operator [] ( unsigned long );
  char operator [] ( unsigned long ) const;
  char & operator [] ( unsigned int );
  char operator [] ( unsigned int ) const;
  char & operator [] ( int );
  char operator [] ( int ) const;
  char & operator [] ( long int );
  char operator [] ( long int ) const;

  // opertors
  // -----------

  operator bool () const;

  //! Converts the LString into a const char* . If LString is currently NULL it returns an empty string.
  operator const char * () const;
  operator const char * ();

  //! Converts the LString into a char* . If LString is currently NULL it returns NULL!
  operator char * ();
  operator char () const;

  operator short unsigned int () const;
  operator short signed int () const;

  operator signed int () const;
  operator unsigned int () const;

  operator signed long () const;
  operator unsigned long () const;

  operator lint8 () const;
  operator luint8 () const;

  //  template < class any_string > operator any_string () const { return any_string ( toCharConst() ); };

  operator double () const;
#ifndef _WIN32
  operator std::string () const;
#endif

  // Queries 
  // -----------

  unsigned long length () const;              //!< Get the length of the string in characters
  char get ( unsigned long ) const;           //!< Get the char at position (const)
  unsigned long contains ( char ) const;      //!< Checks it the string contains a character
  unsigned long numOf ( char ) const;         //!< Same as contains
  bool beginsWith ( LString const &, bool caseSens = false ) const; //!< Check if a string starts with a string
  bool endsWith ( LString const &, bool caseSens = false ) const; //!< Check if a string end with a string


  // Characteristics
  // ------------------

  /*!
   * \name Characteristics
   */
  //@{
  bool isOfFunc ( int (*)( int ) ) const;     //!< Check all characters of this string match a user defined function
  bool isLowerCase () const;                  //!< Check if all characters are lower case
  bool isUpperCase () const;                  //!< Check if all characters are upper case
  bool isDigits () const;                     //!< Check if all characters are digits
  bool isBinaryDigits () const;               //!< Check if all characters are binary digits
  bool isHexDigits () const;                  //!< Check if all characters are hex digits
  bool isAlphabetic () const;
  bool isAlphanumeric () const;               //!< Check if all characters are alphanumeric
  bool isPrintable () const;                  //!< Check if all characters are printable
  bool isControl () const;                    //!< Check if all characters are control characters
  //@}

  // Search and Replace
  // --------------------

  /*!
   * \name Search and Replace
   */
  //@{
  void change ( LString const & scanFor, LString const & replaceWith, unsigned long nums=0 );
  LString searchAndReplace ( LString const & regexp, LString const & ) const; //!< Reg-Exp Search and replace
  //@}

  // chars
  // -----------

  /*!
   * \name Character Processing
   */
  //@{
  char firstChar () const;   //!< Get the first character of a string
  char lastChar () const;    //!< Get the last character of a string

  LString & removeLeading ( LString const & );    //!< Remove all characters leading this string
  LString & removeTrailing ( LString const & );   //!< Remove all characters trailing this string
  
  //! Removes the last character from a string
  LString & removeLastChar ();

  //! Removes the first character from a string
  LString & removeFirstChar ();

  //! Returns a copy of the current string converted to upper case characters
  LString toUpper () const;
  //! Returns a copy of the current string converted to lower case characters
  LString toLower () const;
  //@}

  // convertors (to)
  // -----------

  char * toChar ();            //!< convert this to a char *
  char const * toCharConst () const; //!< convert this to a char const *
  long toLong () const;        //!< convert this to a long
  double toDouble () const;    //!< convert this to a double


  // convertors (from)
  // -----------

  static LString fromDouble ( double value, long acc = -1, long len = -1, char filler = 0 ); //!< Create a string from a double

  static LString fromPrintf ( luint32 maxlen, char const * format, ... );

  // cutters
  // -----------

  /*!
   * \name Cutters
   */
  LString subString ( unsigned long from ) const;  //!< Get a substring from this string (endless)
  LString subString ( unsigned long form, unsigned long len ) const; //!< Get a substring from this string
  LString subString ( unsigned long from, unsigned long len, char ) const; //!< Get a substring from this string (filled up)
  LString left ( unsigned long num ) const; //!< Get the left side of a string
  LString right ( unsigned long num ) const; //!< Get the right side of a string


  // manipulation
  // ---------------

  /*!
   * \name Manipulation
   */
  //@{
  void append ( const char * p1 );    //!< Append a string to this
  void append ( LString const & p1 );    //!< Append a string to this
  void prepend ( LString const & p1 );   //!< Prepand a string to this
  void prepend ( const char * p1 );   //!< Prepand a string to this
  //@}

  // Words 
  // -----------

  /*!
   * \name Word Processing
   */
  //@{
  void setDelims ( const char * );      //!< Set the delimters for word processing
  void setQMarks ( const char * );      //!< Set the quotation marks for word processing
  unsigned long numWords () const;      //!< Get the number of words this string contains
  LString word ( unsigned long ) const; //!< Get a word from this string
  bool containsWord ( LString const & word, bool caseIns = false ) const; //!< Check if the string contains a word
  //@}


  // OS Related
  // -----------

  void getEnv ( const char * var ); //!< Fill this string from the environment


  // Statics
  // -----------

  /*!
   * \name Statics
   */
  //@{
  static LString fromLastError( unsigned long = ::getLastError() );
  static LString perror ( int = errno );        //!< Get the error description from a error code (UNIX)
  static LString fromEnv ( const char * env );  //!< Get a string from the environment
  #ifdef _LCL_SYS_WIN32_
  static LString fromErrorCode ( DWORD rc = ::GetLastError() ); //!< Get the error description from a error code (Win32)
  #endif
  static LString filled ( unsigned long, char ); //!< Get a string filled with a character
  //@}

  // CoDecs
  // -----------

  /*!
   * \name CoDecs
   */
  //@{
  LString deHex();
  LString enHex();
  LString enSave ( LString const & = LString() ) const;
  LString deSave () const;
  //@}

  // Parsing
  // ------------

  LHash<LString,LString> parseToHash ( LString const & regexp ) const;  //!< Build a hash from a string using reg-exps

  // Replacement class
  // ---------------------

  /*!
   * \name Replacing
   */
  //@{

  /*!
   * Default expression for having LCL default (Makefile style) variables   
   *
   * Use with LString::replace
   */
  static const char * C_tag_exp_default;
  /*!
   * Default expression for having BASH like variables.
   *
   * Use with LString::replace
   */
  static const char * C_tag_exp_bash;


  LString replaceVars ( LHash<LString, LString> const &, bool = false ) const;  //!< Replace variables in the string
  LString replaceWithEnvs ( ) const; //!< Replace variables with strings fetched from the environment

  LString replace ( Replacer const & replacer ) const;  //!< Generic replacer
  LString replace_XXX ( Replacer const & ) const;  //!< Generic replacer (to be dumped)
  LString replace ( Replacer const & replacer, LString const & _regEx, luint32 maxrec = 100 ) const; //!< Generic replacer with regexp pattern
 private:
  static LString doReplace ( LString str, Replacer const &, LRegEx & exp, luint32 currec );
 public:
  
  class LCL_EXPORT Replacer {
    friend class LCL_EXPORT LString;
  protected:
    virtual LString sys_replace ( LString const & tag ) const;
  public:
    virtual LString replace ( LString const & tag ) const = 0;

    class LCL_EXPORT TagNotFound {
    public:
      TagNotFound() {};
    };
  };
  
  //@}
};

#include <lcl/liostream.h>

LCL_EXPORT NAMESPACE_STD(ostream) & operator << ( NAMESPACE_STD(ostream) &, LString const & );
LCL_EXPORT LOStream & operator << ( LOStream &, LString const & );
LCL_EXPORT LString operator + ( const char *, LString const & );
LCL_EXPORT LString operator + ( const char *, LString & );
LCL_EXPORT LString operator + ( char *, LString const & );
LCL_EXPORT LString operator + ( char *, LString & );

#include <lcl/lserialize.h>

LCL_EXPORT LSerializer & operator << ( LSerializer & ser, LString const & );
LCL_EXPORT LSerializer & operator >> ( LSerializer & ser, LString & );

/**
 * A convertor to a QT QString
 *
 * \param s any LString
 * \return a QString object
 */
#define asQString(s) (QString((char const*)(s)))

/**
 * A convertor from a QT QString
 *
 * \param s any QString
 * \return a LString object
 */

#define fromQString(s) (LString((char const*)(s)))

/**
 * A convertor to a MFC CString
 *
 * \param s any LString
 * \return a CString object
 */
#define toCString(s) (CString((char const*)(s)))

/**
 * A convertor from a MFC CString
 *
 * \param s any CString
 * \return a LString object
 */

#define fromCString(s) (LString((char const*)(s)))

#include <lcl/lbuffer.h>
#include <lcl/lhash.h>

#include <lcl/lregex.h>

#include <lcl/lstringtokenizer.h>

#endif
