/* ****************************************************************************
   |      File: lthread.h
   |      Contains: class definition for LThread
   |      Created: 17.03.1998
   |      Last Changed: 17.03.1998
   |      Version: 0.0
   |         
   |      ToDo: + Just started
   |            + (global) add some real exceptions !!
   |
   ************************************************************************* */

/* ====>> See also the ChangeLog in lthread.cc <<==== */

#ifndef _L_THREAD_H_
#define _L_THREAD_H_

#include <lcl.h>

class LCL_EXPORT LThread;
class LCL_EXPORT LThreadFn;
//class LCastThreadFn;

LCL_EXPORT extern int countLThreadFn;
LCL_EXPORT extern int countLThread;

#ifdef PURE_WIN32

#include <windows.h>
DWORD WINAPI Win32WrapperThreadFunc( LPVOID ptr );
#define LTHREAD_FUNC_WRAPPER DWORD WINAPI Win32WrapperThreadFunc( LPVOID ptr )

#else

#include <pthread.h>
void * LThreadStarter ( void * );
#define LTHREAD_FUNC_WRAPPER void * LThreadStarter ( void * )

#endif

#include <lcl/lobject.h>
#include <lcl/lvcs.h>
#include <lcl/lxcpt.h>

/*!
 * \defgroup Threads Threads
 */


/*!
 * \class LThread
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A thread control class
 * \ingroup Threads
 *
 * A thread control object for starting new threads
 *
 * 
 */

class LCL_EXPORT LThread : LObject {
  LVCS_DECLARE;

 protected:
#ifdef PURE_WIN32
  HANDLE tid;
#else
  pthread_t tid;
#endif
  LThreadFn * threadFunc;

 public:

  LThread ( LThread const & );
  LThread ( LThreadFn * );
  virtual ~LThread();

  static LThreadID currentID();

  void run (void);

  void start ( bool joinable = false ); //!< Start the thread
  void stop();  //!< Stop the thread
  void kill();  //!< Kill the thread

  void * waitFor(); //!< Wait for the end of the thread

  LThreadFn * getFunc () {return threadFunc;};

  class LCL_EXPORT NotJoinableException : public LBasicException {
  protected:
    LThreadID threadID;
  public:
    NotJoinableException ( char const * _file, luint32 _line, LThreadID );
    LThreadID getThreadID () const;
    virtual void log ( void ) const;
  };
};

#include <lcl/ltrace.h>

/*!
 * \class LThreadFn
 * \brief A thread function class
 * \author Jens Reimann <ctron@dentrassi.de>
 * \ingroup Threads
 *
 * This class is the implementation of a thread function which can be started up in a new thread.
 *
 * To create a new thread function derive a class from LThreadFn and define the run() method.
 *
 * If you don't want to derive you thread function from LThreadFn refer to LFuncThreadFn and LMethodThreadFn
 *
 * None of the LThreadFn classes directly starts a new thread. The only provide a standard interface for LThread.
 * Refer to LThread to see how a thread is started.
 * 
 */

class LCL_EXPORT LThreadFn : LObject {
  LVCS_DECLARE;

  friend class LThread;
  friend LTHREAD_FUNC_WRAPPER;

 protected:  
  bool deleteMe;
  virtual void run () {};

 public:
  LThreadFn() {
    deleteMe = false;
    countLThreadFn++;
  };
  ~LThreadFn () {
    countLThreadFn--;
  }
};

/*!
 * \class LFuncThreadFn
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A wrapper class for threads using a simple C function
 * \ingroup Threads
 *
 * This class provides a simple method of creating a new thread using a simple C function.
 */

class LCL_EXPORT LFuncThreadFn : public LThreadFn {
  LVCS_DECLARE;

  void (*func) ( void ); //!< The function pointer

 protected:  
  virtual void run () {
    if ( func )
      func ();
  };
 public:
  /*!
   * The constructor
   *
   * \param _func The function pointer which sould be started as a thread
   */
  LFuncThreadFn ( void (*_func) (void) ) :
    func ( _func )
    {
      
    }
};

template <class M>
class LCL_EXPORT LCastThreadFn : public LThreadFn {
  LVCS_DECLARE;
 protected:
  M threadData;
  
  virtual void run() {
    run ( threadData );
  };

  virtual void run ( M data ) {
  };

 public:
  LCastThreadFn ( M data ) : LThreadFn(), threadData(data) {};
};


/*!
 * \class LMethodThreadFn
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A wrapper class for threads using a method as new thread
 * \ingroup Threads
 *
 * This class provides a simple method of creating a new thread using a method.
 */

template <class Object>
class LCL_EXPORT LMethodThreadFn : public LThreadFn {
  LVCS_DECLARE;

 protected:
  Object & object; //!< The object which sould be used for the method
  void (Object::*runner) ( void); //!< The method function pointer

  virtual void run () {    

    LTRACE_FUNC ( "LMethodThreadFn", "run" );

    if ( runner )
      (object.*runner) ();
    else
      LTRACE ( "LMethodThreadFn", "runner is NULL" );

  }

 public:
  /*!
   * The constructor
   *
   * \param func The method function pointer
   * \param _object The object which should be used for the function pointer
   */
  LMethodThreadFn ( void (Object::*func)( void ), Object & _object ) :
    LThreadFn (),
    object ( _object ),
    runner ( func )
    {}
};

template <class Object, class P1 >
class LCL_EXPORT LMethodCastThreadFn : public LThreadFn {
  LVCS_DECLARE;

 protected:
  Object & object;
  void (Object::*runner) ( P1 );
  P1 p1;

  virtual void run () {    
    if ( runner )
      (object.*runner) ( p1 );
  }
 public:
  LMethodCastThreadFn ( void (Object::*func)( P1 ), Object & _object, P1 _p1 ) :
    LThreadFn (),
    object ( _object ),
    runner ( func ),
    p1 ( _p1 )
    {}
};

template <class Object, class P1, class P2 >
class LCL_EXPORT LMethodCastThreadFn2 : public LThreadFn {
  LVCS_DECLARE;

 protected:
  Object & object;
  void (Object::*runner) ( P1, P2 );
  P1 p1;
  P2 p2;

  virtual void run () {    
    if ( runner )
      (object.*runner) ( p1, p2 );
  }
 public:
  LMethodCastThreadFn2 ( void (Object::*func)( P1, P2 ), Object & _object, P1 _p1, P2 _p2 ) :
    LThreadFn (),
    object ( _object ),
    runner ( func ),
    p1 ( _p1 ),
    p2 ( _p2 )    
    {}
};



#endif
