#ifndef _LTRANSMITTER_H_
#define _LTRANSMITTER_H_

#include <lpointer.h>
#include <lsockaddr.4inet.h>
#include <lsocket.h>
#include <lstring.h>
#include <lbuffer.h>
#include <lthread.h>
#include <lset.h>
#include <lmap.h>
#include <lsserver.h>

class LTransmitter;
class LTransmitterRequest;
class LTransmitterChannel;
class LTransmitterHandler;
class LTransmitterProtocol;

class LTransmitter {
  friend class LTransmitterChannel_UDP_Server;
  friend class LTransmitterChannel_TCP_Server;
  friend class LTransmitterChannel;

 protected:

  void receivedRequest ( LManagedPointer<LTransmitterRequest> request );
  void receivedAkn ( LManagedPointer<LTransmitterRequest> request );
  void receivedNak ( LManagedPointer<LTransmitterRequest> request );

  void removeInvalidChannels ();

  LMap<LManagedPointer<LTransmitterChannel>, LString > channelList;
  LSet<LManagedPointer<LTransmitterHandler> > handlerList;
  LMap<LManagedPointer<LTransmitterRequest>, unsigned long> requestList;
  //  LHash<unsigned long, LManagedPointer<LTransmitterRequest> > requestList;

  LMethodThreadFn<LTransmitter> checkerFn;

  void addRequest ( LManagedPointer<LTransmitterRequest> );
  void removeRequest ( LManagedPointer<LTransmitterRequest> );

 public:

  LTransmitter ();

  void start ();

  void threadChecker ( void );

  void addChannel ( LManagedPointer<LTransmitterChannel> );
  void removeChannel ( LString const & );

  void addHandler ( LManagedPointer<LTransmitterHandler> );

  void sendRequest ( LManagedPointer<LTransmitterRequest> );
  void aknRequest ( LManagedPointer<LTransmitterRequest> request );
  void nakRequest ( LManagedPointer<LTransmitterRequest> request );

  LEmitter1 < LString > eventInvalidChannel;
  LEmitter1 < LManagedPointer<LTransmitterRequest> > statusChanged;

};

class LTransmitterRequest {
  friend unsigned long key ( LTransmitterRequest const & );
  friend unsigned long key ( LManagedPointer<LTransmitterRequest> );
  friend class LTransmitter;
  friend class LTransmitterProtocol;

 public:
  enum Status { INVALID, INIT, SENT, INBOUND, OUTBOUND, AKN, NAK, TIMEOUT };

 protected:
  Status state;
  LBuffer request, response;
  unsigned long id;
  time_t status_change;
  LString channel_id;

 public:

  LTransmitterRequest ();
  LTransmitterRequest ( LBuffer const &, LString const & channel_id );
  LTransmitterRequest ( LTransmitterRequest const & p1 );

  void setRequestData ( LBuffer const & );
  void setResponseData ( LBuffer const & );
  LBuffer getRequestData ( ) const;
  LBuffer getResponseData ( ) const;

  void setStatus ( Status );
  Status getStatus () const;
  bool isValid () const;

  void setID ( unsigned long );
  unsigned long getID () const;

  void setChannelID ( LString const & id );
  LString getChannelID () const;
};

unsigned long key ( LTransmitterRequest const & );
unsigned long key ( LManagedPointer<LTransmitterRequest> );

class LTransmitterChannel {
  friend class LTransmitterProtocol;  
  friend class LTransmitter;
  friend LString key ( LTransmitterChannel const & );
  friend LString key ( LManagedPointer<LTransmitterChannel> p1 );

 protected:

  LString id;
  LManagedPointer<LTransmitterProtocol> protocol;
  LTransmitter * transmitter;

  virtual void sendPacket ( LBuffer const & ) = 0;

  virtual void check ( void ) {};
  
 public:

  LTransmitterChannel ( LString const & _id, LManagedPointer<LTransmitterProtocol> protocol );

  virtual void receivedData ( LBuffer const & );

  virtual void sendRequest ( LManagedPointer<LTransmitterRequest> );
  virtual void aknRequest ( LManagedPointer<LTransmitterRequest> );
  virtual void nakRequest ( LManagedPointer<LTransmitterRequest> );

  virtual bool isValid () const = 0;

  virtual void requestReceived ( LManagedPointer<LTransmitterRequest> request );
};

LString key ( LTransmitterChannel const & );
LString key ( LManagedPointer<LTransmitterChannel> );

// UDP server
class LTransmitterChannel_UDP_Server : public LTransmitterChannel {
 protected:
  LSocket4Inet socket;

  virtual void sendPacket ( LBuffer const & );

 public:
  LTransmitterChannel_UDP_Server ( LManagedPointer<LTransmitterProtocol> protocol, LSockAddr4Inet const & addr = LSockAddr4Inet::any() );
  virtual bool isValid () const;

  void readEvent ( LFileDesc * );
};

// UDP client
class LTransmitterChannel_UDP_Client : public LTransmitterChannel {
 protected:
  LSocket4Inet socket;
  LSockAddr4Inet addr;
  time_t last_input;

  virtual void sendPacket ( LBuffer const & );

 public:
  LTransmitterChannel_UDP_Client ( LManagedPointer<LTransmitterProtocol> protocol, LSockAddr4Inet const & addr );
  virtual bool isValid () const;

  void readEvent ( LFileDesc * );

  virtual void check ( void );
};

// TCP Client (connect)
class LTransmitterChannel_TCP_Client : public LTransmitterChannel {
 protected:
  LSocket4Inet socket;
  LSockAddr4Inet addr;
  bool connected;
  LBuffer outbound;
  time_t last_input;

 public:

  LTransmitterChannel_TCP_Client ( LManagedPointer<LTransmitterProtocol> protocol, LSockAddr4Inet const & _addr );
  LTransmitterChannel_TCP_Client ( LManagedPointer<LTransmitterProtocol> protocol, LSocket4Inet & _socket );

  virtual void sendPacket ( LBuffer const & buffer );
  virtual void writeEvent ( LFileDesc * fd );
  virtual void readEvent ( LFileDesc * fd  );

  virtual bool isValid () const;
  virtual void check ( void );
};

// TCP Client (connect/reconnect)

class LTransmitterChannel_TCP_Client_AutoCon : public LTransmitterChannel {
 protected:
  LSocket4Inet socket;
  LSockAddr4Inet addr;
  bool connected;
  LBuffer outbound;
  time_t last_input;
  LSlotConnection slot1, slot2;

 public:

  LTransmitterChannel_TCP_Client_AutoCon ( LManagedPointer<LTransmitterProtocol> protocol, LSockAddr4Inet const & _addr );
  LTransmitterChannel_TCP_Client_AutoCon ( LManagedPointer<LTransmitterProtocol> protocol, LSocket4Inet & _socket );

  virtual void sendPacket ( LBuffer const & buffer );
  virtual void writeEvent ( LFileDesc * fd );
  virtual void readEvent ( LFileDesc * fd  );

  virtual bool isValid () const;
  virtual void check ( void );

  void setupSocket ();
  void destroySocket ();
};



// TCP Server (bind+listen)
class LTransmitterChannel_TCP_Server : public LTransmitterChannel {
 protected:
  LSocketServer socketServer;

 public:

  LTransmitterChannel_TCP_Server ( LManagedPointer<LTransmitterProtocol> protocol, LSockAddr4Inet const & _addr );

  virtual void sendPacket ( LBuffer const & buffer );
  virtual void newConnection ( LSocket4Inet & socket );

  virtual bool isValid () const;
};


// Handles incoming requests
class LTransmitterHandler {
 protected:
 public:
  virtual void process ( LManagedPointer<LTransmitterRequest> request, LTransmitter * ) = 0;
};

// Virtual Protocol Class
class LTransmitterProtocol {
 protected:

 public:
  virtual LBuffer encodeRequest ( LManagedPointer<LTransmitterRequest> ) = 0;
  virtual LBuffer encodeAkn ( LManagedPointer<LTransmitterRequest> ) = 0;
  virtual LBuffer encodeNak ( LManagedPointer<LTransmitterRequest> ) = 0;

  virtual void receivedData ( LBuffer const & ) = 0;

  LEmitter1<LManagedPointer<LTransmitterRequest> > requestReceived;

  virtual LManagedPointer<LTransmitterProtocol> clone () = 0;
};

// Default Protocol
class LTransmitterProtocol_default : public LTransmitterProtocol {
 protected:
  LBuffer inbound;

 public:
  virtual LBuffer encodeRequest ( LManagedPointer<LTransmitterRequest> );
  virtual LBuffer encodeAkn ( LManagedPointer<LTransmitterRequest> );
  virtual LBuffer encodeNak ( LManagedPointer<LTransmitterRequest> );

  virtual void receivedData ( LBuffer const  & );

  virtual LManagedPointer<LTransmitterProtocol> clone ();
};

#endif
