/* ****************************************************************************
   |      File: lconfsys.cc
   |      Contains: base class LConfigSystem
   |      Created: 19.03.1998
   |      Last Changed: 19.03.1998
   |      Version: 0.0
   |         
   |      ToDo: + just started
   |
   ************************************************************************* */

/* ************************************************************************* *
 * Change log:                                                               *
 * -------------                                                             *
 *                                                                           *
 *   -> [v0.00] - 19.03.1998 - Created file and started coding               * 
 *   -> [v0.00] - 19.03.1998 - Added all major functions                     *
 *   -> [v0.00] - 21.03.1998 - Found a nasty NULL pointer bug                *
 *   -> [v0.00] - 03.09.1998 - Changed to new List classes                   *
 * ************************************************************************* */


// **** LCL Includes

#include <lcl.h>
#include <lcl/lconfsys.h>
#include <lcl/lpointer.h>
#include <lcl/literator.h>
#include <lcl/ltrace.h>
#include <lcl/lxcpt.h>

LConfigObject defaultObject;

class CountType : public LIterator<LConfigObject> {
public:
  LString type;
  unsigned long counter;

  CountType ( LString str )
    {
      type = str;
      counter = 0;
    }

  virtual void applyTo ( LConfigObject & obj )
    {
      if (obj.isOfType(type)) {
	counter++;
      }
    }
};

class FilterType2 : public LIterator<LConfigObject> {
public:
  LString member;
  LString value;
  LConfigObject * _obj;
  bool casesensitive;

  void applyTo ( LConfigObject & obj )
    {
      if ( obj.isSet ( member ) ) {

	if ( casesensitive ) {

	  if ( obj.get ( "member" ) == value ) {
	    if ( !_obj ) _obj = &obj;
	  }

	} else {

	  if ( obj.get ( "member").toUpper() == value ) {
	    if ( !_obj ) _obj = &obj;
	  }

	} // case

      } // has member
    } // applyTo
};

class FilterType1 : public LIterator<LConfigObject> {
public:
  LString type;
  LIterator<LConfigObject> & iterator;

  FilterType1 ( LString str, LIterator<LConfigObject> & it ) :
    iterator ( it )
    {
      type = str;
    }

  virtual void applyTo ( LConfigObject & obj )
    {
      if (obj.isOfType(type)) {
	iterator.applyTo ( obj );
      }
    }
};

class IterType2 : public LIterator<LConfigObject> {
public:
  LString type;
  void * ptr;
  void (*func)(LConfigObject&, void*);
  void applyTo ( LConfigObject & obj )
    {
      if ( type.length () ) {
	if ( obj.isOfType ( type ) ) {
	  func ( obj, ptr );
	}
      } else {
	func ( obj, ptr );
      }

    }
};

/*!
 * \warning This constructer is depricated! Use the one with LConfigReader instead!
 *
 * Creates a new instance of this class. The data is filled in from the parser passed as argument. The parser will be destroyed after reading in the data.
 * \param parser A pointer to the config parser (e.g. LDefaultConfigParser). The instance
 * of this parser will be deleted after reading in all data
 */
LConfigSystem::LConfigSystem ( LConfigParser * parser )
{
  if ( !parser )
    return;

  parser->parseHeader ( this );
  parser->parseMain ( this );
  delete parser;
}

LConfigSystem::LConfigSystem ( LConfigReader const & reader )
{
  reader.read ( *this );
}

/*!
 * Create an empty config system
 */
LConfigSystem::LConfigSystem ( )
{
}

LConfigSystem::~LConfigSystem ()
{
}

void LConfigSystem::read ( LConfigParser * parser )
{
  if ( !parser )
    throw LNullPointerException();

  parser->parseHeader ( this );
  parser->parseMain ( this );
  delete parser;
}

void LConfigSystem::clear ()
{
  objList.clear ();
}

void LConfigSystem::addObject ( LConfigObject & obj )
{
  objList.add ( obj );
}

/*!
 * Counts all object of a given type
 * \param type The type which will be counted
 * \return the number of items found for this type
 */
unsigned long LConfigSystem::count ( LString const & type )
{
  CountType countEach ( type );
  objList.allElementsDo ( countEach );
  return countEach.counter;
}

/*!
 * Iterates through all objects using a c-func iterator. For each object read in the
 * iterator function is called once. It will get a reference to the LConfigObject and a pointer
 * which may point to application specific data.
 *
 * \param func the pointer to the function
 * \param ptr a pointer which will be passed to the iterator function
 */
void LConfigSystem::forEach ( void (*func)(LConfigObject&, void*), void * ptr )
{
  IterType2 it2;
  it2.func = func;
  it2.ptr = ptr;

  objList.allElementsDo ( it2 );
}

/*!
 * Iterates through all objects using a LIterator object. For each object read in the
 * iterator object is called once. It will get a reference to the LConfigObject.
 *
 * \param iterator refernce to the iterator used
 */
void LConfigSystem::forEach ( LIterator< LConfigObject > & iterator )
{
  objList.allElementsDo ( iterator );
}

/*!
 * Same a forEach (for c-funcs) but it will only iterate through objects of a specified type.
 *
 * \param type specifies the type which will be filtered. The type name is case sensitive!
 * \param func the pointer to the function
 * \param ptr a pointer which will be passed to the iterator function
 */
void LConfigSystem::forEachType ( LString const & type, void (*func)(LConfigObject&, void*), void* ptr)
{
  IterType2 it2;
  it2.func = func;
  it2.ptr = ptr;
  it2.type = type;

  objList.allElementsDo ( it2 );
}

/*!
 * Same a forEach (for LIterator) but it will only iterate through objects of a specified type.
 *
 * \param type specifies the type which will be filtered. The type name is case sensitive!
 * \param iterator refernce to the iterator used
 */
void LConfigSystem::forEachType ( LString const & type, LIterator<LConfigObject> & iterator )
{
  FilterType1 filterType1 ( type, iterator );
  objList.allElementsDo ( filterType1  );
}

/*!
 * Returns the first object of the given type. This is usefull when you exactly know that only
 * one object of the type will exists (e.g. LCGIParser has only one "query" object). If the
 * object does not exists a reference to a static default object is returned. This object is
 * empty.
 *
 * \param type specifies to type to look for
 */
LConfigObject & LConfigSystem::first ( LString const & type )
{
  LManagedPointer<LCursor> cursor ( objList.createCursor () );

  for (cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
    if ( objList.elementAt ( cursor ).isOfType ( type ) )
      return objList.elementAt (cursor);
  }

  return defaultObject;
}

bool LConfigSystem::containsObjectWithMemberEqual ( LString const & _type, LString const & _member, LString const & _value, bool casesensitive ) 
{
  LManagedPointer<LCursor> cursor ( objList.createCursor () );

  LString value = ( casesensitive ? _value : _value.toUpper() );

  for (cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
    
    if ( objList.elementAt ( cursor ).isOfType ( _type ) )
      if ( casesensitive ) {
	if ( objList.elementAt ( cursor ).get ( _member ) == value )
	  return true;
      } else {
	if ( objList.elementAt ( cursor ).get ( _member ).toUpper() == value.toUpper() ) {
	  return true;
	}
      }
    //    return objList.elementAt (cursor);
  }
  
  return false;
}

LConfigObject & LConfigSystem::firstObjectWithMemberEqual ( LString const & _type, LString const & _member, LString const & _value, bool casesensitive )
{

  LManagedPointer<LCursor> cursor ( objList.createCursor () );

  LString value = ( casesensitive ? _value : _value.toUpper() );

  for (cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {

    LTRACE ( "LConfigSystem", "Type: " + objList.elementAt ( cursor ).getType() + " Member: " + objList.elementAt ( cursor ). get ( _member ) + " Value: " + value );

    if ( objList.elementAt ( cursor ).isOfType ( _type ) )
      if ( casesensitive ) {
	if ( objList.elementAt ( cursor ).get ( _member ) == value )
	  return objList.elementAt ( cursor );
      } else {
	if ( objList.elementAt ( cursor ).get ( _member ).toUpper() == value ) {
	  return objList.elementAt ( cursor );
	}
      }
	  

    //      return objList.elementAt (cursor);
  }
  
  return defaultObject;
}



