#include <lcl/ldbsystem.h>

#include <lcl/lstring.h>
#include <lcl/lregex.h>
#include <lcl/ltrace.h>
#include <lcl/llogsystem.h>
#include <lcl/ldirectory.h>
#include <lcl/ldbsystem.plugin.h>
#include <lcl/lpluginloader.h>

#include <lcl/lcl_config.h>

#include <lcl/lrevision.h>

LFEATURE("LCLDB", lcldb_feat1, "PlugIn Support" );
LREVISION("LCLDB", lcldb_rev10, "1.0", "Initial Release" );
LREVISION("LCLDB", lcldb_rev20, "2.0", "Replaced Initial release" );
LREVISION("LCLDB", lcldb_rev22, "2.2", "Added PlugIn Support" );

LTRACE_CHANNEL_STATIC("SQL", __lcl_trace_sql, "Dumping all executed SQL statements");
LTRACE_CHANNEL_STATIC("LDBDriver", __lcl_trace_dbdriver, "Database driver manager information" );
LTRACE_CHANNEL_STATIC("LDBDriver.func", __lcl_trace_dbdriver_func, "LDBDriver function calls" );

bool LDBDriver::loaded_plugins = false;
LHash<LString,LDBDriver_PlugIn*> LDBDriver::driverHash;

void LDBDriver::registerDriver ( LString const & name, LDBDriver_PlugIn * driver_plugin )
{
  LTRACE ( "LDBDriver", "Registered driver: " + name );
  driverHash.set ( name, driver_plugin );
}

class DBPlugLoader : public LPlugInLoader < LDBDriver_PlugIn > {
public:
  void foundPlugIn ( LDBDriver_PlugIn * plugin )
  {
    LDBDriver::registerDriver ( plugin->getName(), plugin ); 
  }
};

void LDBDriver::loadPlugIns ()
{
  loaded_plugins = true;

  /*
  LHeap<LString> list = LDirectory::getList( "/lcl/plugins" );
  LManagedPointer<LCursor> cursor = list.createCursor();

  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      LTRACE ( "LDBDriver", "Loading PlugIn " "/lcl/plugins/" + list.elementAt ( cursor ) );
      LDBDriver_PlugIn * plugin = new LDBDriver_PlugIn ( "/lcl/plugins/" + list.elementAt ( cursor ) );
      if ( plugin->isValid() )
	{
	  LTRACE ( "LDBDriver", "Loaded PlugIn " + plugin->getName() + " -> " + plugin->getDescription() );
	  registerDriver ( plugin->getName(), plugin ); 
	}
      else
	{
	  LTRACE ( "LDBDriver", "PlugIn /lcl/plugins/" + list.elementAt ( cursor ) + " is not a valid DB Driver plug-in" );
	  delete plugin;
	}
    }
  */
  DBPlugLoader loader;
  loader.scan ();
}

LDBDriver_PlugIn * LDBDriver::getDriver ( LString const & name )
{
  if ( !driverHash.isSet ( name )  )
    {
      LTRACE ( "LDBDriver", name + " is not registered!" );
    }
  return driverHash.get ( name , NULL );
}


#define LDBDRIVER(s, d)  if ( db_driver==s ) return new d (configHash) ; LTRACE("LDBDriver", "Registered database driver: " s );

#include <lcl/ldbsystem.includes.h>

/*
 * Database URI in the following form
 *   lcldb[db]://user:passwd@database?additional=options&for=us
 */

/*!
 * \param uri a database URI
 * \return a new database connection or NULL if the creation failed
 *
 * the function expects a database URI in the form of:
 *
 * lcldb[driver_name]://user:password@database
 *
 * or
 *
 * lcldb[driver_name]://user:password@database?option1=value1&option2=value2
 *
 * \sample
 * lcldb[MySQL]://web:secret@system_db?host=databasehost.dentrassi.de
 */
LCL_EXPORT LDBDriver * openDatabase ( LString const & uri )
{

  LDBDriver::loadPlugIns ();

  LTRACE_FUNC ( "LDBDriver", "openDatabase" );
  LRegEx exp ( "^lcldb\\[(.*)\\]://([a-zA-Z0-9\\-_]*):?([a-zA-Z0-9\\-_]*)@([a-zA-Z0-9\\-_]*)\\?\?(.*)$" );

  if ( !exp.exec ( uri ) ) {
    LLOG ( LLogSystem::ERR, "Not a valid LCL Database URI format!" );
    LTRACE ( "LDBDriver", "URI is '" + uri + "'" );
    return NULL;
  }
 
  LString db_driver = exp.match(1);

  LHash<LString,LString> configHash;
  configHash.set ( "DRIVER", exp.match(1) );
  configHash.set ( "USER", exp.match(2) );
  configHash.set ( "PASSWORD", exp.match(3) );
  configHash.set ( "DATABASE", exp.match(4) );

  LString options = exp.match(5);
  LString pair;

  if ( options.length() ) {
    options.setDelims ( "&" );
    for ( unsigned long i = 0; i<options.numWords(); i++) {
      pair = options.word(i);
      pair.setDelims ( "=" );
      configHash.set ( pair.word(0).toUpper(), pair.word(1) );
    }
  }

  #include <lcl/ldbsystem.drivers.h>

  LDBDriver_PlugIn * plugin = LDBDriver::getDriver ( db_driver );

  if ( plugin )
    {
      LDBDriver * driver = plugin->createObject ( configHash );
      if ( !driver )
	LLOG ( LLogSystem::ERR, "Unable to create driver from plugin" );
      return driver;
    }

  LLOG ( LLogSystem::ERR, "Could not find a valid database driver: " + db_driver );

  return NULL;
}

LDBDriver::LDBDriver ( LHash<LString,LString> const & vars ) :
  configHash ( vars )
{
  LTRACE_FUNC ( "LDBDriver", "LDBDriver" );
}

LDBDriver::~LDBDriver ()
{
  LTRACE_FUNC ( "LDBDriver", "~LDBDriver" );
}

/*!
 * \return the user which is specified for the database connection
 */

LString LDBDriver::getUser () const
{
  return getOption ( "USER" );
}

/*!
 * \return the password which is specified for the database connection
 */

LString LDBDriver::getPassword () const
{
  return getOption ( "PASSWORD" );
}

/*!
 * \return the database name which is specified for the database connection
 */

LString LDBDriver::getDatabase () const
{
  return getOption ( "DATABASE" );
}

/*!
 * \return the driver name which is specified for the database connection
 */

LString LDBDriver::getDriverName() const
{
  return getOption ( "DRIVER" );
}

/*!
 * \return the value of the requested option
 *
 * Return the specified value for the database connection of the requested option. Normally you
 * are able to ask options here which are not covered by getUser, getPassword, getDatabase or
 * getDriverName
 */

LString LDBDriver::getOption (LString const & var ) const
{
  LString str;
  try {
    str = configHash.get ( var );
  }
  catch ( ... ) {
  }
  return str;
}

/*!
 * \return a managed pointer to the result of the query
 * \param sql the SQL query which should be executed by the database
 *
 * First an new query handle will be created by the system. Then the SQL command will be executed.
 * The result will be handled by the newly created instance of LDBQuery.
 *
 * You might get a NULL pointer back, so you have to check this befor accessing the result! This may be checked
 * using LManagedPointer::isValid
 */

LManagedPointer<LDBQuery> LDBDriver::query ( LString const & sql )
{
  LTRACE_FUNC ( "LDBDriver", "query" );
  LManagedPointer<LDBQuery> _query = createQuery ();

  if ( _query.isValid() )
    _query->exec ( sql );

  return _query;
}

LDBQuery::LDBQuery ( LDBDriver & db )
{
  valid = false;
}

LDBQuery::LDBQuery ( LDBDriver & db, LString const & _sql )
{
  valid = false;
}

LDBQuery::~LDBQuery ()
{
}

/*!
 * \return \b true if the result set is valid (validated by the database driver)
 *
 * This function will return \b true if the current result set is valid! This value depends on the database
 * driver. So each database (MySQL, ODBC, Oracle...) might handle this status differently.
 *
 * For more information see the specific driver.
 */

bool LDBQuery::isValid () const
{
  return valid;
}

/*!
 * \return will always return true
 * \param _sql the SQL query which should be prepared
 *
 * A SQL query will be prepared for execution. You may call exec afterwards to actually execute the SQL query
 * You can also replace variables using LDBQuery::exec ( LHash<LString, LString> const & hash )
 *
 * This method is only provided for a ODBC like handling of the database connection. It is
 * recommended to use LDBDriver::query instead of this method
 */

bool LDBQuery::prepare ( LString const & _sql )
{
  m_sql = _sql;
  return true;
}

/*!
 * \return \b true if the execution had been successfull.
 *
 * This command will actually execute the query.
 *
 * Normally this method will be called automatically be LDBDriver::query
 *
 * This method is only provided for a ODBC like handling of the database connection. It is
 * recommended to use LDBDriver::query instead of this method
 */

bool LDBQuery::exec ( )
{
  return exec ( m_sql );
}

/*!
 * \return \b true if the execution had been successfull.
 * \param hash the hash table which will be used for variable expansion
 *
 * This command will actually execute the query. But before executing it, it will replace all
 * variables in the SQL query using LString::replaceVars
 *
 * This method is only provided for a ODBC like handling of the database connection. It is
 * recommended to use LDBDriver::query instead of this method
 */

bool LDBQuery::exec ( LHash<LString, LString> const & hash )
{
  LString temp = m_sql.replaceVars ( hash );
  return exec ( temp );
}

LManagedPointer<LDBBind> LDBQuery::createBind ( LString const & )
{
  return NULL;
}
