
#include <lcl.h>

#include <lcl/lfdlayer.posix.h>

#include <lcl/ltrace.h>
#include <lcl/lstreamproxy.h>
#include <lcl/lrevision.h>

LFEATURE("LFDManager", fdm_f1, "POSIX Layer" );
#ifdef HAVE_BROKEN_FSTAT
LFEATURE("LFDManager", fdm_f2, "Compiled with broken fstat fix" );
#endif
#ifdef HAVE_SOCKLEN_T
LFEATURE("LFDManager", fdm_f3, "Compiled with socklen_t support" );
#endif
#ifdef SOCK_DATATYPE_VOID
LFEATURE("LFDManager", fdm_f4, "Compiled with void * datatype support" );
#endif
#ifdef SOCK_DATATYPE_CHAR
LFEATURE("LFDManager", fdm_f5, "Compiled with char * datatype support" );
#endif

LVCS_INIT_D(LFDManager, "POSIX");


LFDManager * LFDManager::theInstance = NULL;
LMutex LFDManager::instanceLock;

LFDManager * LFDManager::getInstance ()
{
  instanceLock.lock ();
  if ( !theInstance )
    theInstance = new LFDManager ();
  instanceLock.unlock ();

  return theInstance;
}

LFDManager::LFDManager ()
{
  lout << "LFDManager -> Creating new LCL File Descriptor Manager for POSIX!" << endl << flush;
  LTRACE_FUNC ( "LFDManager", "LFDManager" );
  LTRACE ( "LFDManager", "Creating new LCL File Descriptor Manager for POSIX!" );
}

bool LFDManager::is_valid ( LFileDescriptor fd ) const
{
  LTRACE_FUNC("LFDManager", "is_valid");
  return fd != LINVALID_FILEDESC;
}

bool LFDManager::is_socket ( LFileDescriptor fd ) const
{
  LTRACE_FUNC("LFDManager", "is_socket");
#ifdef HAVE_BROKEN_FSTAT
  int l1;
  int len = sizeof(l1);
  return ::getsockopt ( fd, SOL_SOCKET, SO_DEBUG, (char*)&l1, (L_ADDRLEN_T*)&len ) == 0;

#else
  struct stat si;
  if ( ::fstat ( fd, &si ) == -1 )
    return false;
  return S_ISSOCK ( si.st_mode );
#endif
}

int LFDManager::sys_open ( char const * filename, int accessType, mode_t om )
{
  LTRACE_FUNC("LFDManager", "sys_open");
  int ocf = 0;
  if ( (accessType & read) && !(accessType & write) )
    ocf = O_RDONLY;
  if ( (accessType & write) && !(accessType & read) )
    ocf = O_WRONLY;
  if ( (accessType & read) && (accessType & write) )
    ocf = O_RDWR;
  int _fd;
  if ( om == 0 )
    {
      _fd = ::open ( filename,
			 ocf |
			 ((accessType & create) ? O_CREAT : 0) |
			 ((accessType & truncate) ? O_TRUNC : 0) |
			 ((accessType & failExists) ? O_EXCL : 0) |
			 ((accessType & append) ? O_APPEND : 0)
			 );
    }
  else
    {
      _fd = ::open ( filename,
			 ocf |
			 ((accessType & create) ? O_CREAT : 0) |
			 ((accessType & truncate) ? O_TRUNC : 0) |
			 ((accessType & failExists) ? O_EXCL : 0) |
			 ((accessType & append) ? O_APPEND : 0),
			 om
			 );
    }
  if ( _fd == -1 )
    LTRACE_PERROR();
    
  return _fd;
}

int LFDManager::sys_select ( int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout )
{
  LTRACE_FUNC("LFDManager", "sys_select");
  return ::select ( nfds, readfds, writefds, exceptfds, timeout );
}

int LFDManager::sys_setsockopt ( LFileDescriptor fd, int level, int optname, void const * optval, unsigned long optlen )
{
  LTRACE_FUNC("LFDManager", "sys_setsockopt");
  return ::setsockopt ( fd, level, optname, (L_SOCK_DATA_TYPE_CONST)optval, optlen );
}

int LFDManager::sys_getsockopt ( LFileDescriptor fd, int level, int optname, void * optval, unsigned long *optlen )
{
  LTRACE_FUNC("LFDManager", "sys_getsockopt");
  L_OPTLEN_T _optlen = *optlen;
  int rc = ::getsockopt ( fd, level, optname, (L_SOCK_DATA_TYPE)optval, &_optlen );
  *optlen = _optlen;
  return rc;
}

int LFDManager::sys_getsockname ( LFileDescriptor fd, struct sockaddr *name, unsigned long *namelen )
{
  LTRACE_FUNC("LFDManager", "sys_getsockname");
  L_ADDRLEN_T _addrlen = *namelen;
  int rc = ::getsockname ( fd, name, &_addrlen );
  *namelen = _addrlen;
  return rc;
}

int LFDManager::sys_getpeername ( LFileDescriptor fd, struct sockaddr *name, unsigned long *namelen )
{
  LTRACE_FUNC("LFDManager", "sys_getpeername");
  L_ADDRLEN_T _addrlen = *namelen;
  int rc = ::getpeername ( fd, name, &_addrlen );
  *namelen = _addrlen;
  return rc;
}

int LFDManager::sys_sendto ( LFileDescriptor fd, void const * data, unsigned long len, int flags, struct sockaddr const *to, unsigned long tolen )
{
  //  LTRACE_FUNC("LFDManager", "sys_sendto");
  return ::sendto ( fd, (L_SOCK_DATA_TYPE_CONST)data, len, flags, to, tolen );
}

int LFDManager::sys_recvfrom ( LFileDescriptor fd, void * data, unsigned long len, int flags, struct sockaddr *from, unsigned long * fromlen )
{
  LTRACE_FUNC("LFDManager", "sys_recvfrom");
  L_ADDRLEN_T _addrlen = *fromlen;
  int rc = ::recvfrom ( fd, (L_SOCK_DATA_TYPE)data, len, flags, from, &_addrlen );
  *fromlen = _addrlen;
  return rc;
}

LFileDescriptor LFDManager::sys_accept ( LFileDescriptor fd, struct sockaddr * addr, unsigned long * addr_len )
{
  LTRACE_FUNC("LFDManager", "sys_accept");
  L_ADDRLEN_T _addrlen = *addr_len;  
  int rc = ::accept ( fd, addr, &_addrlen );
  *addr_len = _addrlen;
  return rc;
}

int LFDManager::sys_bind ( LFileDescriptor fd, struct sockaddr const * addr, unsigned long addr_len )
{
  LTRACE_FUNC("LFDManager", "sys_bind");
 
  #ifdef HAVE_CONST_CONNECT

  return ::bind ( fd, addr, addr_len );

  #else

  return ::bind ( fd, (struct sockaddr*)addr, addr_len );

  #endif
}

int LFDManager::sys_connect ( LFileDescriptor fd, struct sockaddr const * addr, unsigned long addr_len )
{
  LTRACE_FUNC("LFDManager", "sys_connect");

  #ifdef HAVE_CONST_CONNECT

  return ::connect ( fd, addr, addr_len );

  #else

  return ::connect ( fd, (struct sockaddr*)addr, addr_len );

  #endif
}

int LFDManager::sys_listen ( LFileDescriptor fd, int backlog )
{
  LTRACE_FUNC("LFDManager", "sys_listen");
  return ::listen ( fd, backlog );
}

int LFDManager::sys_dup ( LFileDescriptor fd )
{
  LTRACE_FUNC("LFDManager", "sys_dup");
  return ::dup ( fd );
}

int LFDManager::sys_read ( LFileDescriptor fd, void * ptr, unsigned long size )
{
  //  LTRACE_FUNC("LFDManager", "sys_read");
  return ::read ( fd, ptr, size );
}

int LFDManager::sys_write ( LFileDescriptor fd, void const * ptr, unsigned long size )
{
  //  LTRACE_FUNC("LFDManager", "sys_write");
  int rc = ::write ( fd, ptr, size );
  if ( rc == -1 )
    LTRACE_PERROR();
  return rc;
}

int LFDManager::sys_socket ( int af, int type, int proto )
{
  LTRACE_FUNC("LFDManager", "sys_socket");
  return ::socket ( af, type, proto );
}

int LFDManager::sys_close ( LFileDescriptor fd )
{
  LTRACE_FUNC("LFDManager", "sys_close");
  return ::close ( fd );
}

bool LFDManager::sys_getblocking ( LFileDescriptor fd )
{
  int rc = ::fcntl ( fd, F_GETFL ) & O_NONBLOCK;

  if ( rc == -1 )
    return false;

  return rc;
}

void LFDManager::sys_setblocking ( LFileDescriptor fd, bool state )
{
  if ( !state )
    ::fcntl ( fd, F_SETFL, ::fcntl ( fd, F_GETFL ) | O_NONBLOCK );
  else
    ::fcntl ( fd, F_SETFL, ::fcntl ( fd, F_GETFL ) & ~O_NONBLOCK );
}
