#include <lcl.h>

#include <lcl/lfiledesc.h>
#include <lcl/lfdlayer.h>

LVCS_INIT(LFileDesc);

LFileDesc::LFileDesc () :
  fd ( LINVALID_FILEDESC ),
  auto_close ( true )
{
}

LFileDesc::LFileDesc ( LFileDescriptor _fd ) :
  fd ( LINVALID_FILEDESC ),
  auto_close ( true )
{
  assign ( _fd );
}

LFileDesc::LFileDesc ( LFileDesc const & p1 ) :
  fd ( LINVALID_FILEDESC ),
  auto_close ( p1.auto_close )
{
  copyFrom ( p1 );
}

LFileDesc::~LFileDesc ()
{
  if ( auto_close ) {
    if ( isValid () )
      LSelectLoop::unregisterForAll ( this );
    close ();
  }
}

void LFileDesc::assign ( LFileDescriptor _fd )
{
  if ( isValid () )
    close ();

  fd = _fd;
  //  cout << "assign " << fd << endl;
}

/*!
 * Close the file descriptor. The instance will be invalid afterwards but may be reused.
 */

void LFileDesc::close ()
{
  LFDManager::getInstance()->sys_close ( fd );

  fd = LINVALID_FILEDESC;
}

bool LFileDesc::isValid ( ) const 
{
  return LFDManager::getInstance () -> is_valid ( fd );
  //  return fd != LINVALID_FILEDESC;
}

/*!
 * Normally if you destroy your file descriptor instance the corresponding file descriptor
 * will also be closed. If you want your file descriptor to stay open set the auto close flag
 * to \em false
 * \param _flag new value of the auto close flag
 */

void LFileDesc::setAutoClose ( bool _flag )
{
  auto_close = _flag;
}

/*!
 * \param p1 the file descriptor object to copy
 * Copies an existing file descriptor and assigns it to this instance
 */
void LFileDesc::copyFrom ( LFileDesc const & p1 )
{

  LFileDescriptor new_fd = LFDManager::getInstance() -> sys_dup ( p1.getHandle() );

  assign ( new_fd );
}

/*!
 * \return the operating system handle of the file descriptor
 */
LFileDescriptor LFileDesc::getHandle ( ) const 
{
  return fd;
}

void LFileDesc::emitReadEvent ()
{
  readEvent.emit ( this );
}

void LFileDesc::emitWriteEvent ()
{
  writeEvent.emit ( this );
}

void LFileDesc::emitXcptEvent ()
{
  xcptEvent.emit ( this );
}

void LFileDesc::setEventMask ( unsigned long _eMask )
{
  //  eMask = _eMask;
  LSelectLoop::getInstance () -> fdMasks.set ( fd, _eMask );
}

unsigned long LFileDesc::eventMask () const
{
  return LSelectLoop::getInstance () -> fdMasks.get ( fd, 0 );
  //  return eMask;
}

/*!
 * \param loop The LSelectLoop with which this file descriptor should be registered
 *
 * After registering the file descriptor the LSelectLoop instance will report
 * event for this file descriptor based on the event mask set using setEventMask.
 */
void LFileDesc::reg ( LSelectLoop * loop )
{
  loop->reg ( this );
}

/*!
 * \param loop The LSelectLoop from which this file descriptor should be unregistered
 *
 * After unregistering the file descriptor the LSelectLoop instance won't any more emit
 * event for this file descriptor
 */
void LFileDesc::unreg ( LSelectLoop * loop )
{
  loop->unreg ( this );
}

int LFileDesc::write ( void const * ptr, size_t len )
{
  return LFDManager::getInstance() -> sys_write ( fd, ptr, len );
}

int LFileDesc::write ( LBuffer const & buffer )
{
  return write ( buffer.getConstPtr(), buffer.getSize() );
}

int LFileDesc::read ( void * ptr, size_t len )
{
  return LFDManager::getInstance () -> sys_read ( fd, ptr, len );
}

int LFileDesc::read ( LBuffer & _buffer, size_t len )
{
  LBuffer buffer ( len );
  int rc = read ( (void*)buffer, buffer.getSize() );
  _buffer = buffer;
  _buffer.resize ( rc );
  return _buffer.getSize ();
}

#ifndef _WIN32
void LFileDesc::setBlocking ( bool state )
{
  LFDManager::getInstance() -> sys_setblocking ( fd, state );
}

bool LFileDesc::getBlocking () const
{
  return LFDManager::getInstance() -> sys_getblocking ( fd );
}
#endif

LFileDesc & LFileDesc::operator = ( LFileDesc const & p1 )
{
  copyFrom ( p1 );
  return *this;
}

/*
int LFileDesc::fcntl ( int cmd ) const
{
  return ::fcntl ( fd, cmd );
}

int LFileDesc::fcntl ( int cmd, long arg )
{
  return ::fcntl ( fd, cmd, arg );
}
*/
