#include <lcl/lmemory.h>

#include <lcl/lhash.h>
#include <lcl/lmutex.h>
#include <lcl/lutils.h>
#include <lcl/lstreamproxy.h>

#include <lcl/lrevision.h>
LFEATURE("LCL Memory Debug", _lcl_mem_f1, "LCL Memory Counter" );
LREVISION("LCL Memory Debug", _lcl_mem_r1, "1.0", "Initial Version" );
LREVISION("LCL Memory Debug", _lcl_mem_r2, "1.1", "Safer Locking" );

#include <iostream.h>

typedef struct _lcl_memory_heap {
  char * name;
  luint32 blocks;
  luint32 size;
  _lcl_memory_heap * next;
} lcl_memory_heap;

typedef struct _lcl_memory_block {
  luint32 size;
  void * data;
  lcl_memory_heap * heap;
} lcl_memory_block;

static LHash < void*, lcl_memory_block* > blockHash;
lcl_memory_heap * root_heap = NULL;
lcl_memory_heap * last_heap = NULL;

LGuardMutex memoryLock;

LCL_EXPORT bool lmem_debug = false;

inline lcl_memory_heap * lmem_locate_heap ( char const * name )
{
  LAutoLock _lock ( memoryLock );

  lcl_memory_heap * current = root_heap;
  while ( current )
    {
      if ( strcmp ( name, current->name ) == 0 )
	return current;

      current = current->next;
    }
  return NULL;
}

inline lcl_memory_heap * lmem_fetch_heap ( char const * name )
{
  LAutoLock _lock ( memoryLock );

  lcl_memory_heap * heap = lmem_locate_heap ( name );

  if ( !heap )
    {
      heap = new lcl_memory_heap;
      heap->name = strdup ( name );
      heap->size = 0;
      heap->blocks = 0;
      heap->next = NULL;
      
      if ( !root_heap )
	root_heap = heap;
      if ( last_heap )
	last_heap->next = heap;
      last_heap = heap;
    }

  return heap;
}

inline lcl_memory_block * lmem_fetch_block ( void * ptr )
{
  LAutoLock _lock ( memoryLock );

  return blockHash.get ( ptr, NULL );
}

inline bool lmem_validate_ptr ( void * ptr )
{
  LAutoLock _lock ( memoryLock );

  return lmem_fetch_block ( ptr ) != NULL;
}

inline void lmem_free_block ( void * ptr )
{
  LAutoLock _lock ( memoryLock );

  if ( !lmem_validate_ptr ( ptr ) )
    {
      cerr << "LCL Memory: WARNING!! Freeing invalid pointer (" << ptr << ")!" << endl << flush;
      return;
    }
  lcl_memory_block * block = lmem_fetch_block ( ptr );

  blockHash.unset ( block->data );

  free ( block->data );
  block->heap->size -= block->size;
  block->heap->blocks--;
  delete block;
}

inline void * lmem_alloc_block ( luint32 size, char const * name )
{
  LAutoLock _lock ( memoryLock );

  lcl_memory_heap * heap = lmem_fetch_heap ( name );

  lcl_memory_block * block = new lcl_memory_block;
  block->heap = heap;
  block->size = size;
  block->heap->size += block->size;
  block->heap->blocks++;
  block->data = malloc ( size );
  //  cerr << "Allocated new block: " << block->data << endl << flush;

  blockHash.set ( block->data, block );

  return block->data;
}

// Exported functions

LCL_EXPORT void * lmem_malloc ( luint32 size, char const * name )
{
  if ( size == 0 )
    return NULL;

  return lmem_alloc_block ( size, name );
}

LCL_EXPORT void * lmem_realloc ( void * ptr, luint32 size, char const * name )
{
  LAutoLock _lock ( memoryLock );

  lcl_memory_block * block = lmem_fetch_block ( ptr );

  if ( !ptr )
    return lmem_malloc ( size, name );

  if ( size == 0 )
    return NULL;

  if ( !block )
    {
      lerr << "LCL Memory: WARNING! Block should exist!" << endl << flush;
      return lmem_malloc ( size, name );
    }
  else
    {
      void * _ptr = lmem_malloc ( size, block->heap->name );
      if ( block->data )
	{
	  memcpy ( _ptr, block->data, MIN(size,block->size) ); 
	}

      lmem_free_block ( ptr );

      return _ptr;
    }
}

LCL_EXPORT void lmem_free ( void * ptr )
{
  if ( ptr == NULL )
    return;

  lmem_free_block ( ptr );
}

LCL_EXPORT char * lmem_strdup ( char const * str, char const * name )
{

  unsigned long len = strlen ( str );

  void * ptr = lmem_alloc_block ( len+1, name );
  strcpy ( (char*)ptr, str );

  return (char*)ptr;
}

LCL_EXPORT void lmem_dump_stats ( LOStream & out )
{
  out << endl;
  out << "Memory Statistic Dump:" << endl;
  out << "--------------------------------------------------" << endl;

  LAutoLock _lock ( memoryLock );

  lcl_memory_heap * current = root_heap;
  while ( current )
    {
      out << current->name << " : " << current->blocks << " = " << current->size << endl;
      current = current->next;
    }

  out << "--------------------------------------------------" << endl;
  out << flush;
}
