#include <lcl.h>

#include <lcl/lmutex.h>
#include <lcl/lrevision.h>
#include <lcl/ltrace.h>

#include <lcl/lapp.h>
#include <lcl/lstreamproxy.h>

#ifdef NEED_FALLBACK_GUARD_MUTEX
LFEATURE ( "LIPC", __lipc_ge_mutex_f1, "Emulated recursive mutex" );
#endif

#ifdef _LCL_SYS_WIN32_

LVCS_INIT_D ( LMutex, " WIN32" );
LFEATURE ( "LIPC", __lipc_mutex_f1, "Win32 Mutex Lock" );

LMutex::LMutex ()
{  
  mutxp = CreateMutex ( NULL, FALSE, NULL );
}

LMutex::~LMutex ()
{
  CloseHandle ( mutxp );
}

void LMutex::lock () const
{
  WaitForSingleObject ( mutxp, INFINITE );
}

bool LMutex::trylock () const
{
  return WaitForSingleObject ( mutxp, 0 ) == WAIT_OBJECT_0;
}

void LMutex::unlock () const
{
  ReleaseMutex ( mutxp );
}

// LGuardMutex

LGuardMutex::LGuardMutex ()
{
  InitializeCriticalSection ( &crit );
}

LGuardMutex::~LGuardMutex ()
{
  DeleteCriticalSection ( &crit );
}

void LGuardMutex::lock () const
{
  EnterCriticalSection ( (CRITICAL_SECTION *)&crit );
}

void LGuardMutex::unlock () const
{
  LeaveCriticalSection ( (CRITICAL_SECTION *)&crit );
}

bool LGuardMutex::trylock () const
{
  return TryEnterCriticalSection ( (CRITICAL_SECTION *)&crit );
}


#else

LVCS_INIT_D ( LMutex, " Posix" );
LFEATURE ( "LIPC", __lipc_mutex_f1, "POSIX Mutex Lock" );

LMutex::LMutex ()
{
  //  LTRACE_FUNC ( "LMutex", "LMutex" );

  if ( LApplication::debugFlag ) lerr << "LMutex::LMutex()" << endl << flush;

  pthread_mutex_init ( &mutxp, 0 );
}

LMutex::~LMutex ()
{
  //  LTRACE_FUNC ( "LMutex", "LMutex" );

  if ( LApplication::debugFlag ) lerr << "LMutex::~LMutex()" << endl << flush;

  pthread_mutex_destroy ( &mutxp );
}

void LMutex::lock () const
{
  //  LTRACE_FUNC ( "LMutex", "lock" );

  if ( LApplication::debugFlag ) lerr << "LMutex::lock()" << endl << flush;

  pthread_mutex_lock ( (pthread_mutex_t*)&mutxp );
}

bool LMutex::trylock () const
{
  //  LTRACE_FUNC ( "LMutex", "trylock" );

  if ( LApplication::debugFlag ) lerr << "LMutex::trylock()" << endl << flush;

  if ( pthread_mutex_trylock ( (pthread_mutex_t*)&mutxp ) )
    {
      // FIXME: Better error checking!!
      return false;
    }
  else
    {
      return true;
    }

}

void LMutex::unlock () const
{
  //  LTRACE_FUNC ( "LMutex", "unlock" );

  if ( LApplication::debugFlag ) lerr << "LMutex::unlock()" << endl << flush;

  pthread_mutex_unlock ( (pthread_mutex_t*)&mutxp );
}


// LGuardMutex

#ifndef NEED_FALLBACK_GUARD_MUTEX

pthread_mutexattr_t LGuardMutex::guard_attr;
bool LGuardMutex::guard_attr_init = false;
void LGuardMutex::checkInit ()
{
  if ( !guard_attr_init )
    {
      guard_attr_init = true;
      pthread_mutexattr_init ( &guard_attr );

      pthread_mutexattr_settype ( &guard_attr, PTHREAD_MUTEX_RECURSIVE );
    }
}

LGuardMutex::LGuardMutex ()
{
  checkInit ();
  pthread_mutex_init ( &mutxp, &guard_attr );
}

#else

#include <lcl/lfallback_mutex.h>

LGuardMutex::LGuardMutex () :
  mtx ( new __lcl_fallbackGuardMutex )
{
}

LGuardMutex::~LGuardMutex ()
{
  delete mtx;
}

void LGuardMutex::lock () const
{
  mtx->lock ();
}

void LGuardMutex::unlock () const
{
  mtx->unlock ();
}

bool LGuardMutex::trylock () const
{
  return mtx->trylock ();
}
#endif

#endif

/*!
 * \example autolock.cc
 * This example shows you how to use LAutoLock
 */

/*!
 * Creates the autolock instance and locks the mutex (using LMutex::lock)
 */
LAutoLock::LAutoLock ( LMutex const & _mutex ) :
  mutex ( _mutex ),
  m_attached ( true )
{
  //  LTRACE_FUNC ( "LAutoLock", "LAutoLock" );

  mutex.lock ();
}

/*!
 * Unlock the mutex (using LMutex::unlock) and destroy the autolock instance
 */
LAutoLock::~LAutoLock ()
{
  //  LTRACE_FUNC ( "LAutoLock", "~LAutoLock" );

  m_attached = false;
  mutex.unlock ();
}

/*!
 * Detaches from the mutex lock and unlocks it
 *
 * If this is called twice or more, the mutex will not bu unlocked again!
 */
void LAutoLock::detach ()
{
  //  LTRACE_FUNC ( "LAutoLock", "detach" );

  if ( m_attached )
    mutex.unlock ();

  m_attached = false;
}

