#include <lcl/lselectloop.h>

#include <lcl.h>
#include <lcl/lhashiter.h>
#include <lcl/lutils.h>

#include <iostream.h>

#ifndef _WIN32
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#endif

#include <lcl/lstreamproxy.h>

#include <lcl/lfdlayer.h>
#include <lcl/ltrace.h>

class UnregDescIter : public LHashIterator<LThreadID, LSelectLoop*> {
public:
  LFileDesc * fd;
  void applyTo ( LThreadID & tid,  LSelectLoop * & loop ) {
    loop->unreg ( fd );
  }
};

LHash < LThreadID, LSelectLoop* > LSelectLoop::loopThreadHash;

LSelectLoop::LSelectLoop () :
  running ( true ),
  timeoutValue ( 1000 )
{
  //  cout << "LSelectLoop -> New select loop created for " << LGetCurrentThreadID() << "!" << endl; 
}

LSelectLoop * LSelectLoop::getInstance ()
{
  if ( !loopThreadHash.isSet ( LGetCurrentThreadID() ) ) {
    LSelectLoop * newLoop = new LSelectLoop ();
    loopThreadHash.set ( LGetCurrentThreadID(), newLoop );
  }

  return loopThreadHash.get ( LGetCurrentThreadID() );

}

void LSelectLoop::unregisterForAll ( LFileDesc * fd )
{
  LTRACE_FUNC ( "LSelectLoop", "unregisterForAll" );
  UnregDescIter udi;
  udi.fd = fd;
  loopThreadHash.allElementsDo ( udi );
}

#ifdef _WIN32
void LSelectLoop::generatePollList ()
{
  LFileDesc * fd;

  read_list.clear ();
  write_list.clear ();
  except_list.clear ();

  LManagedPointer<LCursor> cursor = fdList.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {

    fd = fdList.elementAt ( cursor );

    if ( fd->eventMask () & LFileDesc::eventRead ) {
      read_list.add ( fd->getHandle() );
    }
    if ( fd->eventMask () & LFileDesc::eventWrite ) {
      write_list.add ( fd->getHandle() );
    }
    if ( fd->eventMask () & LFileDesc::eventXcpt ) {
      except_list.add ( fd->getHandle() );
    }
  }

}

long LSelectLoop::doThePoll ()
{
  if ( LFDManager::getInstance () -> sys_select ( read_list, write_list, except_list, read_res, write_res, except_res, timeoutValue ) )
    {
      return except_res.entries() + write_res.entries() + read_res.entries();
    }
  else
    {
      return -1;
    }
}

void LSelectLoop::emitFDEvents ()
{

  LSet<LFileDesc*> procfdList = fdList;
  LManagedPointer<LCursor> cursor;

  cursor = procfdList.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      LFileDesc * fd = procfdList.elementAt ( cursor );
      if ( read_res.containsElement ( fd->getHandle() ) )
	fd->emitReadEvent ();
      if ( write_res.containsElement ( fd->getHandle () ) )
	fd->emitWriteEvent ();
      if ( except_res.containsElement ( fd->getHandle () ) )
	fd->emitXcptEvent ();
    }
}

#else
void LSelectLoop::generatePollList ( )
{
  LTRACE_FUNC ( "LSelectLoop", "generatePollList" );

  LFileDesc * fd;

  FD_ZERO ( &r_fds );
  FD_ZERO ( &w_fds );
  FD_ZERO ( &x_fds );
  max_fd = -1;
  long num_fds = 0;

  //  lout << "Poll List: ";

  LCursor * cursor = fdList.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
    fd = fdList.elementAt ( cursor );

    if ( fd->isValid () ) {

      if ( fd->eventMask () & LFileDesc::eventRead ) {
	//	lout << ">" << (long)fd->getHandle();
	FD_SET ( fd->getHandle(), &r_fds );
	max_fd = MAX ( max_fd, fd->getHandle() );
	num_fds++;
      }
      if ( fd->eventMask () & LFileDesc::eventWrite ) {
	//	lout << "<" << (long)fd->getHandle();
	FD_SET ( fd->getHandle(), &w_fds );
	max_fd = MAX ( max_fd, fd->getHandle() );
	num_fds++;
      }
      if ( fd->eventMask () & LFileDesc::eventXcpt ) {
	//	lout << "|" << (long)fd->getHandle();
	FD_SET ( fd->getHandle(), &x_fds );
	max_fd = MAX ( max_fd, fd->getHandle() );
	num_fds++;
      }
    }
  }
  //  lout << endl << flush;
  delete cursor;
  max_fd ++;
  // cout << "generated WIN32 fds, max: " << max_fd << " num: " << num_fds << " iil: " << fdList.entries() << endl;
}

long LSelectLoop::doThePoll ()
{
  LTRACE_FUNC ( "LSelectLoop", "doThePoll" );

  struct timeval tv;
  tv.tv_sec = timeoutValue/1000;
  tv.tv_usec = (timeoutValue%1000)*1000;

  //  cout << "pre-select" << endl;

  //  int rc = select ( max_fd, &r_fds, &w_fds, &x_fds, &tv );
  return LFDManager::getInstance() -> sys_select ( max_fd, &r_fds, &w_fds, &x_fds, &tv );

  //  cout << "post-select" << endl;
  /*
  if ( rc == -1 )
    perror ( "select" );

  if ( rc == 0 ) {
    LTRACE ( "LSelectLoop", "Emitting timeout" );
    timeoutEvent.emit ();
  }

  if ( rc > 0 ) {
    emitFDEvents ();
  }
  */
    
}

void LSelectLoop::emitFDEvents ()
{
  LSet<LFileDesc*> procfdList = fdList;

  LFileDesc * fd;
  LCursor * cursor = procfdList.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
    fd = procfdList.elementAt ( cursor );

    if ( fd->isValid() )
    if ( FD_ISSET ( fd->getHandle(), &r_fds ) && (fd->eventMask() | LFileDesc::eventRead ) )
      fd->emitReadEvent ();

    if ( fd->isValid() )
    if ( FD_ISSET ( fd->getHandle(), &w_fds ) && (fd->eventMask() | LFileDesc::eventWrite ) )
      fd->emitWriteEvent ();

    if ( fd->isValid() )
    if ( FD_ISSET ( fd->getHandle(), &x_fds ) && (fd->eventMask() | LFileDesc::eventXcpt ) )
      fd->emitXcptEvent ();
  }
  delete cursor;
  
}

#endif

void LSelectLoop::loop ()
{

  //  cout << "LSelectLoop -> Entering loop for " << LGetCurrentThreadID () << endl;

  while ( running ) {

    generatePollList ();
    long rc = doThePoll ();

    if ( rc == -1 )
      LTRACE_PERROR();
    
    if ( rc == 0 ) {
      //    cout << "emitting timeout" << endl;
      LTRACE ( "LSelectLoop", "Emitting timeout" );
      timeoutEvent.emit ();
    }

    if ( rc > 0 ) {
      //    cout << "emitting fd events" << endl;
      emitFDEvents ();
    }
    
  }
  
}

void LSelectLoop::reg ( LFileDesc * fd )
{
  if ( fdList.add ( fd ) )
    LTRACE ( "LSelectLoop", " Registered handle " + LString( (long)fd->getHandle() ) );
}

void LSelectLoop::unreg ( LFileDesc * fd )
{
  fdList.removeElement ( fd );
  LTRACE ( "LSelectLoop", " UnRegistered handle " + LString( (long)fd->getHandle() ) );
}

void LSelectLoop::setTimeout ( unsigned long _timeoutValue )
{
  timeoutValue = _timeoutValue;
}
