#include <lcl/lsocket.h>

#include <lcl/lfdlayer.h>
#include <lcl/ltrace.h>

LSocket::LSocket ( int _domain, int _type, int _proto ) :
  LFileDesc ( LFDManager::getInstance() -> sys_socket ( _domain, _type, _proto ) ),
  domain ( _domain ),
  type ( _type ),
  protocol ( _proto )
{
}

LSocket::LSocket ( LFileDescriptor _fd ) :
  LFileDesc ( _fd )
{
  if ( !LFDManager::getInstance() -> is_socket ( _fd ) )
    throw LWrongParameterException ( __FILE__, __LINE__, "File Descriptor " + LString(_fd) + " is not socket!" );
}

void LSocket::listen ( long backlog )
{
  if ( LFDManager::getInstance() -> sys_listen ( fd, backlog ) == -1 ) {
    //    perror ( "LSocket::listen" );
    LTRACE_PERROR();
  }
}
  
bool LSocket::connect ( LSockAddr const & _addr )
{
  if ( LFDManager::getInstance() -> sys_connect ( fd, _addr, _addr.length() ) == -1 ) {
    //    perror ( "LSocket::connect" );
    LTRACE_PERROR();
    return false;
  }
  return true;
}
 
void LSocket::bind ( LSockAddr const & _addr )
{
  if ( LFDManager::getInstance () -> sys_bind ( fd, _addr, _addr.length() ) == -1 ) {
    LTRACE_PERROR();
    //    perror ( "LSocket::bind" );
  }
}


void LSocket::reuseaddr ( bool _state )
{
  int state = _state;
  if ( setsockopt ( SOL_SOCKET, SO_REUSEADDR, &state, sizeof(state) ) == -1 ) {
    LTRACE_PERROR();
    //    perror ( "LSocket::reuseaddr" );
  }
}

int LSocket::setsockopt ( int level, int optname, const void *optval, int optlen )
{
  return LFDManager::getInstance () -> sys_setsockopt ( fd, level, optname, optval, optlen );

}

void LSocket::linger ( long _amount )
{
  struct linger li;
  li.l_onoff = _amount ? 1 : 0;
  li.l_linger = _amount;
  if ( setsockopt ( SOL_SOCKET, SO_LINGER, &li, sizeof(li) ) == -1 ) {
    LTRACE_PERROR();
    //    perror ( "LSocket::linger" );
  }
}

int LSocket::sendto ( void const * ptr, int len, sockaddr const * addr, unsigned long addrlen, int flags  )
{
  return LFDManager::getInstance () -> sys_sendto ( fd, ptr, len, flags, addr, addrlen );

}

int LSocket::sendto ( LBuffer const & buffer, LSockAddr const & addr, int flags )
{
  return sendto ( buffer.getConstPtr(), buffer.getSize(), addr, addr.length(), flags );
}

int LSocket::recvfrom ( void * ptr, int len, sockaddr * addr, unsigned long * addrlen, int flags )
{
  return LFDManager::getInstance () -> sys_recvfrom ( fd, ptr, len, flags, addr, addrlen );

}

// ------------>>>>>>>>>>>>>>>>>>> LSocket4Inet

LSocket4Inet::LSocket4Inet ( int _domain, int _type, int _proto ) :
  LSocket ( _domain, _type, _proto )
{
  
}

LSocket4Inet::LSocket4Inet ( LFileDescriptor _fd ) :
  LSocket ( _fd )
{
}


bool LSocket4Inet::isConnected ( LFileDescriptor _fd )
{
  sockaddr_in addr;
  unsigned long len = sizeof(addr);

  int rc = LFDManager::getInstance()->sys_getpeername ( _fd, (sockaddr*)&addr, &len );
  return rc == 0;
}

bool LSocket4Inet::isConnected () const
{
  return isConnected ( fd );
}


LSocket * LSocket4Inet::accept ()
{
  sockaddr_in addr;
  unsigned long addr_len = sizeof ( addr );

  LFileDescriptor new_fd = LFDManager::getInstance () -> sys_accept ( fd, (sockaddr*)&addr, &addr_len );

  if ( !LFDManager::getInstance() -> is_valid ( new_fd ) ) {
    LTRACE_PERROR();
    //    perror ( "LSocket4Inet::accept" );
    return NULL;
  }


  LTRACE ( "LSocket4Inet", "Accepted fd is " + LString(new_fd) );

  return new LSocket4Inet ( new_fd );
}

long LSocket4Inet::getType ()
{
  int i;
  unsigned long ilen = sizeof(i);

  int rc = LFDManager::getInstance () -> sys_getsockopt ( fd, SOL_SOCKET, SO_TYPE, (void*)&i, &ilen );


  if ( rc == -1 ) {
    //    perror ( "LSocket4Inet::getType" );
    LTRACE_PERROR();
    return -1;
  } else {
    return i;
  }
}

LSockAddr4Inet LSocket4Inet::localAddr ( )
{
  sockaddr_in addr;
  unsigned long addr_len = sizeof(addr);
  LFDManager::getInstance () -> sys_getsockname ( fd, (sockaddr*)&addr, &addr_len );

  return LSockAddr4Inet ( &addr );
}

LSockAddr4Inet LSocket4Inet::remoteAddr ( )
{
  sockaddr_in addr;
  unsigned long addr_len = sizeof(addr);
  LFDManager::getInstance () -> sys_getpeername ( fd, (sockaddr*)&addr, &addr_len );

  return LSockAddr4Inet ( &addr );
}

int LSocket4Inet::recvfrom ( LBuffer & buffer, LSockAddr4Inet & addr, int flags )
{
  sockaddr_in _addr;
  unsigned long addr_len = sizeof ( _addr );
  int rc = LSocket::recvfrom ( buffer.getPtr(), buffer.getSize(), (sockaddr*)&_addr, &addr_len, flags );
  addr = LSockAddr4Inet ( &_addr );
  return rc;
}

bool LSocket4Inet::addMembership ( LSockAddr4Inet const & group, LSockAddr4Inet const & addr )
{
  #ifdef PURE_WIN32

  return false;

  #else  

  struct ip_mreq mreq;
  int ret;

 /* add membership */
  mreq.imr_multiaddr = group;
  mreq.imr_interface = addr;
  //  ret = ::setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*) &mreq, sizeof(mreq));
  ret = LFDManager::getInstance() -> sys_setsockopt (fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*) &mreq, sizeof(mreq));

  return ret >= 0;

  #endif
}

// ------------>>>>>>>>>>>>>>>>>>> LSocketUnix
#ifndef _WIN32

LSocketUnix::LSocketUnix ( int _domain, int _type, int _proto ) :
  LSocket ( _domain, _type, _proto )
{
  
}

LSocketUnix::LSocketUnix ( LFileDescriptor _fd ) :
  LSocket ( _fd )
{
}


bool LSocketUnix::isConnected ( LFileDescriptor _fd )
{
  sockaddr_un addr;
  unsigned long len = sizeof(addr);

  int rc = LFDManager::getInstance()->sys_getpeername ( _fd, (sockaddr*)&addr, &len );
  return rc == 0;
}

bool LSocketUnix::isConnected () const
{
  return isConnected ( fd );
}


LSocket * LSocketUnix::accept ()
{
  sockaddr_un addr;
  unsigned long addr_len = sizeof ( addr );

  LFileDescriptor new_fd = LFDManager::getInstance () -> sys_accept ( fd, (sockaddr*)&addr, &addr_len );

  if ( !LFDManager::getInstance() -> is_valid ( new_fd ) ) {
    //    perror ( "LSockAddrUnix::accept" );
    LTRACE_PERROR();
    return NULL;
  }


  LTRACE ( "LSocketUnix", "Accepted fd is " + LString(new_fd) );

  return new LSocketUnix ( new_fd );
}

long LSocketUnix::getType ()
{
  int i;
  unsigned long ilen = sizeof(i);

  int rc = LFDManager::getInstance () -> sys_getsockopt ( fd, SOL_SOCKET, SO_TYPE, (void*)&i, &ilen );


  if ( rc == -1 ) {
    //    perror ( "LSockAddrUnix::getType" );
    LTRACE_PERROR();
    return -1;
  } else {
    return i;
  }
}

LSockAddrUnix LSocketUnix::localAddr ( )
{
  sockaddr_un addr;
  unsigned long addr_len = sizeof(addr);
  LFDManager::getInstance () -> sys_getsockname ( fd, (sockaddr*)&addr, &addr_len );

  return LSockAddrUnix ( &addr );
}

LSockAddrUnix LSocketUnix::remoteAddr ( )
{
  sockaddr_un addr;
  unsigned long addr_len = sizeof(addr);
  LFDManager::getInstance () -> sys_getpeername ( fd, (sockaddr*)&addr, &addr_len );

  return LSockAddrUnix ( &addr );
}

int LSocketUnix::recvfrom ( LBuffer & buffer, LSockAddrUnix & addr, int flags )
{
  sockaddr_un _addr;
  unsigned long addr_len = sizeof ( _addr );
  int rc = LSocket::recvfrom ( buffer.getPtr(), buffer.getSize(), (sockaddr*)&_addr, &addr_len, flags );
  addr = LSockAddrUnix ( &_addr );
  return rc;
}

#endif
