
#include <lcl.h>
#include <lcl/lsserver.h>
#include <lcl/lutils.h>
#include <lcl/lthread.h>
#include <lcl/lemit.h>
#include <lcl/lmutex.h>

#include <lcl/lstreamproxy.h>
#include <lcl/ltrace.h>

#include <lcl/lrevision.h>

#ifdef AF_INET
LFEATURE("LCL Socket Server", _lss_af_inet, "Address Family: AF_INET" );
#endif

#ifdef AF_UNIX
LFEATURE("LCL Socket Server", _lss_af_unix, "Address Family: AF_UNIX" );
#endif

/**
 * \struct ServerAndSocket
 * \internal
 */
typedef struct {
  LSocket4Inet * socket;
  LSocketServer * server;
} ServerAndSocket;

class ConnectionWrapperThread : public LCastThreadFn<ServerAndSocket> {
  CLASSNAME("ConnectionWrapperThread");
  LVCS_DECLARE;

protected:
  virtual void run ( ServerAndSocket sas )
    {
      //      cout << "New thread processing...\n";
      sas.server->connection.emit ( *sas.socket );

      // Socket must be deleted in this thread!!!
      sas.socket->close ();
      delete sas.socket;
      deleteMe = true;
    };

public:
  ConnectionWrapperThread (ServerAndSocket s ) : LCastThreadFn<ServerAndSocket>(s) {};
  
};

LVCS_INIT(ConnectionWrapperThread);
LVCS_INIT(LSocketServer);

LSocketServer::LSocketServer () :
  serverSocket ( NULL ),
  asyncConnections ( true )
{
}

LSocketServer::~LSocketServer ()
{
  destroySocket ();
}

bool LSocketServer::valid () const
{
  return serverSocket != NULL;
}

void LSocketServer::setAsync ( bool flag )
{
  asyncConnections = flag;
}

void LSocketServer::destroySocket ()
{
  delete serverSocket;
  serverSocket = NULL;
}

void LSocketServer::createSocket ( SocketOptions opts, SocketDomain domain, SocketType type, SocketProtocol protocol )
{
  // first destroy old socket
  destroySocket ();

  int socketType, socketDomain, socketProtocol;

  switch ( domain ) {

#ifdef AF_INET
  case LSS_AF_INET:
    socketDomain = AF_INET;
    break;
#endif

#ifdef AF_UNIX
  case LSS_AF_UNIX:
    socketDomain = AF_UNIX;
    break;
#endif

#ifdef AF_ISO    
  case LSS_AF_ISO:
    socketDomain = AF_ISO;
    break;
#endif

#ifdef AF_NS
  case LSS_AF_NS:
    socketDomain = AF_NS;
    break;
#endif

#ifdef AF_IMP
  case LSS_AF_IMP:
    socketDomain = AF_IMP;
    break;
#endif

  default:
    socketDomain = -1;
    break;
  }

  switch ( type ) {
  case LSS_TYPE_STREAM:
    socketType = SOCK_STREAM;
    break;    
  case LSS_TYPE_DGRAM:
    socketType = SOCK_DGRAM;
    break;
  case LSS_TYPE_RAW:
    socketType = SOCK_RAW;
    break;
  case LSS_TYPE_SEQPACKET:
    socketType = SOCK_SEQPACKET;
    break;
  case LSS_TYPE_RDM:
    socketType = SOCK_RDM;
    break;
  default:
    socketDomain = -1;
    break;
  }

  switch ( protocol ) {
  case LSS_PROT_DEFAULT:
    socketProtocol = 0;
    break;
  case LSS_PROT_IP:
    socketProtocol = GetProtocolByName ( "IP" );
    break;
  case LSS_PROT_ICMP:
    socketProtocol = GetProtocolByName ( "ICMP" );
    break;
  case LSS_PROT_IGMP:
    socketProtocol = GetProtocolByName ( "IGMP" );
    break;
  case LSS_PROT_GGP:
    socketProtocol = GetProtocolByName ( "GGP" );
    break;
  case LSS_PROT_TCP:
    socketProtocol = GetProtocolByName ( "TCP" );
    break;
  case LSS_PROT_PUP:
    socketProtocol = GetProtocolByName ( "PUP" );
    break;
  case LSS_PROT_UDP:
    socketProtocol = GetProtocolByName ( "UDP" );
    break;
  case LSS_PROT_IDP:
    socketProtocol = GetProtocolByName ( "IDP" );
    break;
  default:
    socketProtocol = -1;
    break;
  }

  if ( socketDomain == -1 )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "unkown domain" );
    }
  if ( socketType == -1 )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "unkown socket type" );
    }
  if ( socketProtocol == -1 )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "unkown protocol" );
    }
    

  // creating new socket
  serverSocket = new LSocket4Inet ( socketDomain, socketType, socketProtocol );

  // setting socket options
  if ( opts & reuseaddr )
    serverSocket->reuseaddr ();

  // connection socket events
  connect_emitter_1 ( &LSocketServer::serverSocketRead, *this, serverSocket->readEvent );
}

void LSocketServer::bind ( LSockAddr4Inet const & addr )
{
  if ( !valid() )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "no socket created" );
    }

  serverSocket->bind ( addr );
}

void LSocketServer::listen ( int backlog )
{
  if ( !valid() )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "no socket created" );
    }

  serverSocket->listen ( backlog );
}


#ifndef _WIN32
#warning Currently only SOCK_STREAM and SOCK_DGRAM connections are supported by LSocketServer
#endif

void LSocketServer::serverSocketRead ( LFileDesc * fd )
{
  if ( !valid() )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "no socket created" );
    }
  //    throw 1;

  if ( fd != serverSocket )
    return;

  LSocket4Inet & socket = *(LSocket4Inet*)fd;

  if ( socket.getType() == SOCK_DGRAM ) {
    #ifndef _WIN32
    #warning Have to make packetBuffer size variable
    #endif
    LBuffer packetBuffer ( 1024 * 10 );
    //    LIP ip;
    LSockAddr4Inet ip;
    
    socket.recvfrom ( packetBuffer, ip, 0);

    packet.emit ( packetBuffer, ip);

  } // END -> SOCK_DGRAM

  if ( socket.getType() == SOCK_STREAM ) {

    //    lout << "Accepting connection!" << endl << flush;
    LTRACE ( "LSocketServer", "Accpeting connection" );
    LSocket4Inet * clientSocket = (LSocket4Inet*)socket.accept ();
    LTRACE ( "LSocketServer", "New connection fd=" + LString(clientSocket->getHandle()) );
    emitConnection ( clientSocket );

  } // END -> SOCK_STREAM

}

void LSocketServer::emitConnection ( LSocket4Inet * clientSocket )
{
  if ( asyncConnections ) {
    ServerAndSocket sas;
    
    sas.server = this;
    sas.socket = clientSocket;
    
    // create thread...
    LThread * clientThread = new LThread ( new ConnectionWrapperThread ( sas ) );
    
    // launch new thread
    clientThread->start ();
    
    // may be deleted now because the thread_fn runs without it
    delete clientThread;

  } else {
    connection.emit ( *clientSocket );
    clientSocket->close ();
  }

}

void LSocketServer::start ()
{
  if ( !valid () )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "no socket created" );
    }
  //    throw 1;
  
  //  serverSocket->setEventMask ( LFileDesc::read );
  serverSocket->setEventMask ( LFileDesc::eventRead );
  serverSocket->reg ( LSelectLoop::getInstance () );

  startServer.emit ( *this );
}

void LSocketServer::stop ()
{
  if ( !valid () )
    {
      throw LWrongParameterException ( __FILE__, __LINE__, "no socket created" );
    }
  //    throw 1;

  //  serverSocket->setEventMask ( 0 );
  serverSocket->setEventMask ( 0 );
  serverSocket->unreg ( LSelectLoop::getInstance () );

  stopServer.emit ( *this );
}

LSocket4Inet & LSocketServer::getSocket ()
{
  return *serverSocket;
}

LSocket4Inet const & LSocketServer::getSocket () const
{
  return *serverSocket;
}
