#include <lcl.h>

#include <lcl/ltrace.h>
#include <lcl/lstreamproxy.h>
#include <lcl/ldatetime.h>
#include <lcl/ltracechannel.h>
#include <lcl/lregex.h>
#include <lcl/lapp.h>
#include <lcl/lthread.h>

#include <string.h>
#include <time.h>

LTraceMessage::LTraceMessage () :
  m_channel ( NULL ),
  m_msg ( NULL ),
  m_file ( NULL ),
  m_line ( 0L ),
  m_time ( 0L ),
  m_utime ( 0L ),
  m_pid ( 0L ),
  m_tid ( 0L )
{
}

LTraceMessage::LTraceMessage ( LTraceMessage const & p1 ) :
  m_channel ( strdup ( p1.m_channel ) ),
  m_msg ( strdup ( p1.m_msg ) ),
  m_file ( strdup ( p1.m_file ) ),
  m_line ( p1.m_line ),
  m_time ( p1.m_time ),
  m_utime ( p1.m_utime ),
  m_pid ( p1.m_pid ),
  m_tid ( p1.m_tid )
{
}

LTraceMessage::LTraceMessage ( char const * channel, char const * msg, char const * _file, luint32 _line, luint32 _time, luint32 _utime, luint32 _pid, luint32 _tid ) :
  m_channel ( strdup ( channel ) ),
  m_msg ( strdup ( msg ) ),
  m_file ( strdup ( _file ) ),
  m_line ( _line ),
  m_time ( _time ),
  m_utime ( _utime ),
  m_pid ( _pid ),
  m_tid ( _tid )
{
}

LTraceMessage::~LTraceMessage ()
{
  if ( m_channel )
    free ( m_channel );
  if ( m_msg )
    free ( m_msg );
  if ( m_file )
    free ( m_file );
}

char const * LTraceMessage::getChannel () const
{
  return m_channel ? m_channel : "";
}

char const * LTraceMessage::getMessage () const
{
  return m_msg ? m_msg : "";
}

char const * LTraceMessage::getFile () const
{
  return m_file ? m_file : "";
}

luint32 LTraceMessage::getTime() const
{
  return m_time;
}

luint32 LTraceMessage::getUTime () const
{
  return m_utime;
}

luint32 LTraceMessage::getLine () const
{
  return m_line;
}

luint32 LTraceMessage::getPID () const
{
  return m_pid;
}

luint32 LTraceMessage::getTID () const
{
  return m_tid;
}

class __lcl_tracex_match : public LHashIterator < LString, luint32 > {
public:
  bool match;
  LString channel;

  void applyTo ( LString & _key, luint32 & count )
  {
    if ( !match && (count>0) )
      {
	LRegEx exp ( _key );
	if ( exp.exec ( channel ) )
	  match = true;
      }
  };
};

class __lcl_tracex_trace : public LHashIterator < LString, LManagedPointer < LTraceOutput > > {
public:
  LTraceMessage * msg;
  void applyTo ( LString & _key, LManagedPointer < LTraceOutput > & output )
  {
    output->trace ( *msg );
  }
};

LCL_EXPORT LTracerX * LTracerX::theInstance = NULL;
LCL_EXPORT LGuardMutex LTracerX::instanceLock;
LTracerX * LTracerX::getInstance ()
{
  LAutoLock _lock ( instanceLock );

  if ( !theInstance )
    theInstance = new LTracerX;

  return theInstance;
}

LTracerX::LTracerX () :
  enabled ( false )
{
  lerr << "LCL -> Created LTracer..." << endl << flush;
}

void LTracerX::trace ( LTraceMessage const & msg )
{
  setInTrace ( true );

  LDEBUG ( "Tracing for channel: " + LString(msg.getChannel()) ) ;

  __lcl_tracex_trace tracerx;
  tracerx.msg = (LTraceMessage*)&msg;

  controlLock.lock ();
  outputHash.allElementsDo ( tracerx );  
  controlLock.unlock ();

  // Now process stickies ...

  setInTrace ( false );
}

bool LTracerX::isMatch ( LString const & channel )
{
  __lcl_tracex_match matcher;
  matcher.match = false;
  matcher.channel = channel;
  
  controlLock.lock ();
  filterList.allElementsDo ( matcher );
  controlLock.unlock ();

  if ( matcher.match )
    {
      LDEBUG ( "Channel: " + channel + " matched!" );
    }
  else
    {
      LDEBUG ( "Channel: " + channel + " did not match!" );
    }

  return matcher.match;
}

void LTracerX::addChannel ( LString const & channel )
{
  controlLock.lock ();
  if ( !channelStates.isSet ( channel ) )
    {
      channelStates.set ( channel, isMatch ( channel ) );
    }
  controlLock.unlock ();
}

bool LTracerX::channelActive ( LString const & channel )
{
  addChannel ( channel );

  LAutoLock _lock ( controlLock );

  return channelStates.get ( channel, false );

}

void LTracerX::calculateStates ( void )
{
  controlLock.lock ();

  LSet<LString> channels = channelStates.getKeys ();
  LCursor * cursor = channels.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      channelStates.set ( channels.elementAt ( cursor ), isMatch ( channels.elementAt ( cursor ) ) );
    }
  delete cursor;
  controlLock.unlock ();
}

void LTracerX::addFilter ( LString const & filter )
{
  controlLock.lock ();
  luint32 count = filterList.get ( filter, 0 );
  filterList.set ( filter, count + 1 );
  if ( !count )
    calculateStates ();
  controlLock.unlock ();
}

void LTracerX::removeFilter ( LString const & filter )
{
  controlLock.lock ();
  luint32 count = filterList.get ( filter, 0 );
  filterList.set ( filter, MAX ( count - 1, 0 ) );
  //  lout << filterList.get ( filter, 0 ) << "<<--" << endl;
  if ( count == 1 )
    calculateStates ();
  controlLock.unlock ();
}

void LTracerX::setInTrace ( bool state )
{
  controlLock.lock ();
  recursiveStates.set ( LGetCurrentThreadID(), state );
  controlLock.unlock ();
}

bool LTracerX::inTrace ( void ) const
{
  bool state;
  controlLock.lock ();
  state = recursiveStates.get ( LGetCurrentThreadID(), false );
  controlLock.unlock ();
  return state;
}

void LTracerX::addOutput ( LString const & name, LManagedPointer<LTraceOutput> output )
{
  controlLock.lock ();
  if ( outputHash.isSet ( name ) )
    {
      outputHash.get ( name ) -> unreg ();
    }
  outputHash.set ( name, output );
  output->reg ();
  controlLock.unlock ();
}

void LTracerX::removeOutput ( LString const & name )
{
  controlLock.lock ();
  if ( outputHash.isSet ( name ) )
    {
      outputHash.get ( name ) -> unreg ();
      outputHash.unset ( name );
    }
  controlLock.unlock ();
}

LSet<LString> LTracerX::getOutputs () const
{
  LAutoLock _lock ( controlLock );
  return outputHash.getKeys ();
}

LManagedPointer<LTraceOutput> LTracerX::getOutput ( LString const & name )
{
  LAutoLock _lock ( controlLock );
  return outputHash.get ( name, NULL );
}

// Temporarily removed! Need a clear concept!!!
/*
LHash<LString, luint32> & LTracerX::getStickiesHash ()
{
  LAutoLock _lock ( controlLock );
  if ( !stickies.isSet ( LGetCurrentThreadID() ) )
    {
      stickies.set ( LGetCurrentThreadID(), LHash<LString, luint32> );
    }
  return stickies.get ( LGetCurrentThreadID() );
}

void LTracerX::addSticky ( LString const & sticky )
{
  LHash<LString, luint32> hash = getStickiesHash();
  luint32 value = 0;
  if ( hash.isSet ( sticky ) )
    {
      value = hash.get ( sticky );
    }
  value++;
  hash.set ( sticky, value );
}

void LTracerX::removeSticky ( LString const & sticky )
{
  LHash<LString, luint32> hash = getStickiesHash();
  luint32 value = 0;
  if ( hash.isSet ( sticky ) )
    {
      value = hash.get ( sticky );
      value--;
    }  
  if ( value )
    {
      hash.set ( sticky, value );
    }
  else
    {
      hash.unset ( sticky );
    }
}
*/

// LTraceOutput


LTraceOutput::LTraceOutput () :
  isReg ( false )
{  
}

LTraceOutput::~LTraceOutput ()
{
}

void LTraceOutput::reg ()
{
  controlLock.lock ();
  if ( !isReg )
    {
      isReg = true;
      LCursor * cursor = filterList.createCursor ();
      forcursor ( cursor )
	{
	  LTracerX::getInstance() -> addFilter ( filterList.elementAt ( cursor ) );
	}
      delete cursor;
    }
  controlLock.unlock ();
}

void LTraceOutput::unreg ()
{
  controlLock.lock ();
  if ( isReg )
    {
      isReg = false;
      LCursor * cursor = filterList.createCursor ();
      forcursor ( cursor )
	{
	  LTracerX::getInstance() -> removeFilter ( filterList.elementAt ( cursor ) );
	}
      delete cursor;
    }
  controlLock.unlock ();
}

void LTraceOutput::addFilter ( LString const & filter )
{
  controlLock.lock ();
  if ( !filterList.containsElement ( filter ) )
    {
      filterList.add ( filter );
      if ( isReg )
	LTracerX::getInstance() -> addFilter ( filter );
    }
  controlLock.unlock ();
}

void LTraceOutput::removeFilter ( LString const & filter )
{
  controlLock.lock ();
  if ( filterList.containsElement ( filter ) )
    {
      filterList.removeElement ( filter );
      if ( isReg )
	LTracerX::getInstance() -> removeFilter ( filter );
    }
  controlLock.unlock ();
}

void LTraceOutput::traceX ( LTraceMessage const & msg )
{
  if ( isMatch ( msg ) )
    {
      trace ( msg );
    }
}

bool LTraceOutput::isMatch ( LTraceMessage const & msg )
{
  LAutoLock _lock ( controlLock );

  LCursor * cursor = filterList.createCursor ();
  LString channel = msg.getChannel ();
  forcursor ( cursor )
    {
      LRegEx exp ( filterList.elementAt ( cursor ) );
      if ( exp.exec ( channel ) )
	{
	  delete cursor;
	  return true;	  
	}
    }
  delete cursor;
  return false;
}

LSet<LString> LTraceOutput::getFilters () const
{
  LAutoLock _lock ( controlLock );
  return filterList;
}

// Automatic Channel Lists

static LTraceChannelInfo * _channelInfoHead = NULL;
static LTraceChannelInfo * _channelInfoTail = NULL;

__lcl_traceChannelInfo::__lcl_traceChannelInfo ( char const * name, char const * filter, char const * desc )
{
  if ( !name )
    return;

  LTraceChannelInfo * temp = new LTraceChannelInfo;
  temp->next = NULL;
  temp->name = strdup ( name );
  if ( filter )
    temp->filter = strdup ( filter );
  else
    temp->filter = NULL;
  if ( desc )
    temp->desc = strdup ( desc );
  else
    temp->desc = NULL;

  if ( _channelInfoTail )
    {
      _channelInfoTail->next = temp;
      _channelInfoTail = temp;
    }
  else
    {
      _channelInfoHead = _channelInfoTail = temp;
    }
}

LCL_EXPORT void __lcl_DumpTraceChannels ( LOStream & ostream )
{
  ostream << "-------------------- LCL Registered Trace Channels -------------------" << endl;
  LTraceChannelInfo * current = _channelInfoHead;
  while ( current ) {
    ostream << current->name << " | ";
    if ( current->filter )
      {
	ostream << current->filter;
      }
    else
      {
	ostream << current->name;
      }
    ostream << " | ";
    if ( current->desc )
      {
	ostream << current->desc;
      }
    else
      {
	ostream << "<no description>";
      }
    ostream << endl;
    current = current->next;
  }
  ostream << "--------------------   End of Trace Channel Dump   -------------------" << endl;
  ostream << flush;  
}
