#include <ldbsystem.odbc.h>

// should not be built anyway
//#ifdef HAVE_LIBODBC

#include <lcl/ltrace.h>
#include <lcl/lplugin.h>

LDBDriver_ODBC::LDBDriver_ODBC ( LHash<LString,LString> const & vars ) :
  LDBDriver ( vars )
{
  LTRACE_FUNC ( "LDBDriver_ODBC", "LDBDriver_ODBC" );
  valid = false;

  connect ();
}

LDBDriver_ODBC::~LDBDriver_ODBC ()
{
  LTRACE_FUNC ( "LDBDriver_ODBC", "~LDBDriver_ODBC" );
  disconnect ();
}

LManagedPointer<LDBQuery> LDBDriver_ODBC::createQuery ()
{
  LTRACE_FUNC ( "LDBDriver_ODBC", "createQuery" );
  return new LDBQuery_ODBC ( *this );
}

bool LDBDriver_ODBC::connect ()
{
  LTRACE_FUNC ( "LDBDriver_ODBC", "connect" );

  if ( connected () )
    return false;

  valid = false;

  //  cout << getOption("HOST") << getUser() <<  getPassword() << getDatabase() << endl;

  LTRACE ( "LDBDriver_ODBC", "SQLConnect: Database/User/Password = " + getDatabase() + "/" + getUser() + "/" + getPassword() );

  if ( SQLAllocEnv ( &henv ) != SQL_SUCCESS  ) {
    LTRACE ( "LDBDriver_ODBC", "SQLAllocEnv failed: " + getLastODBCError() );
    return false;
  }

  if ( SQLAllocConnect ( henv, &hdbc ) != SQL_SUCCESS  ) {
    LTRACE ( "LDBDriver_ODBC", "SQLAllocConnect failed: " + getLastODBCError() );
    return false;
  }

  if ( SQLConnect ( hdbc, (unsigned char*)(char*)getDatabase(), getDatabase().length(), (unsigned char*)(char*)getUser(), getUser().length(), (unsigned char*)(char*)getPassword(), getPassword().length() ) != SQL_SUCCESS ) {
    LTRACE ( "LDBDriver_ODBC", "SQLConnect failed: " + getLastODBCError() );
    return false;
  }

  valid = true;

  LTRACE ( "LDBDriver_ODBC", "Connected to database" );
  
  return connected();
}

bool LDBDriver_ODBC::disconnect ()
{
  if ( !connected() )
    return false;

  SQLDisconnect ( hdbc );
  SQLFreeConnect ( hdbc );
  SQLFreeEnv ( henv );

  valid = false;

  return true;
}

bool LDBDriver_ODBC::connected () const
{
  return valid;
}


LDBQuery_ODBC::LDBQuery_ODBC ( LDBDriver_ODBC & drv ) :
  LDBQuery ( drv ),
  driver ( drv )
{
  SQLAllocStmt ( driver.hdbc, &hstmt );
  valid = false;
  rowCount = 0;
}

LDBQuery_ODBC::LDBQuery_ODBC ( LDBDriver_ODBC & drv, LString const & sql ) :
  LDBQuery ( drv, sql ), 
  driver ( drv )
{
  SQLAllocStmt ( driver.hdbc, &hstmt );
  valid = false;
  rowCount = 0;
}

LDBQuery_ODBC::~LDBQuery_ODBC ()
{
  if (hstmt)
    SQLFreeStmt ( hstmt, SQL_DROP );
  hstmt = NULL;
  valid = false;
}

bool LDBQuery_ODBC::hasResult () const
{
  if ( !hstmt )
    return false;

  return numRows();
}

long LDBQuery_ODBC::numFields () const
{
  if ( !valid || !hasResult() )
    return 0;
  
  SWORD cnt;
  
  if ( SQLNumResultCols ( hstmt, &cnt ) != SQL_SUCCESS ) {
    LTRACE ( "LDBQuery_ODBC", "SQLNumResultCols failed: " + getLastODBCError() );
    return 0;
  }
  
  return cnt;  
}

long LDBQuery_ODBC::numRows () const
{
  if ( !valid )
    return 0;

  return rowCount;  
}

bool LDBQuery_ODBC::exec ( LString const & sql )
{
  LString _sql ( sql );

  SQLPrepare ( hstmt, (unsigned char*)(char*)_sql, _sql.length() );

  valid = ((SQLExecute( hstmt ) == SQL_SUCCESS) ? true:false);


  LTRACE ( "SQL", "Executing: '" + sql + "'");

  if (!valid) {
    LString buffer = getLastODBCError ();
    //    cout << "[SQLError] for : '" << sql << "'" << endl;
    //    cout << buffer << endl << flush;
    LTRACE ( "SQLErrors", "Query '" + sql + "' failed! Reason: '" + buffer + "'" );
    return false;
  }


  unsigned long cnt = 0;

  _sql.setDelims ( " \n" );
  LTRACE ( "LDBSystem_ODBC", "Basic SQL cmd is '" + _sql.word(0).toUpper() + "'" );
  if ( _sql.word ( 0 ).toUpper() == "SELECT" ) {
    // Count all rows ... this is cruel but the only way to get the number of
    // rows returned by an SQL SELECT statment
    LTRACE ( "LDBSystem_ODBC", "Counting rows for select statement!" );
       
    LHash<LString,LString> tempHash;
    
    while ( fetchNext(tempHash) ) {
      tempHash.clear ();
      cnt++;
    } 
    SQLFreeStmt ( hstmt, SQL_DROP );
    SQLAllocStmt ( driver.hdbc, &hstmt );
    SQLPrepare ( hstmt, (unsigned char*)(char*)_sql, _sql.length() );
    if ( SQLExecute ( hstmt ) != ERROR_SUCCESS ) {
      LTRACE ( "SQLErrors", "Query '" + sql + "' failed! Reason: '" + getLastODBCError() + "'" );
    }
  }

  rowCount = cnt;

  return isValid();
}

bool LDBQuery_ODBC::fetchNext ()
{
  internalHash.clear();
  return fetchNext ( internalHash );
}

bool LDBQuery_ODBC::fetchNext ( LHash<LString, LString> & hash )
{
  long i;

  if ( !isValid() )
    return false;

  if ( SQLFetch ( hstmt ) != SQL_SUCCESS )
    return false;

  for ( i = 0; i<numFields(); i++ ) {
    hash.set ( getFieldName ( i ), sys_getField ( i ) );
  }

  return true;
}

LString LDBQuery_ODBC::sys_getField ( long i ) const
{
  char buffer[4*1024];

  if ( !isValid() || !hasResult() )
    return LString();

  if ( i < 0 )
    return LString();

  SDWORD len;

  if ( i < numFields() ) {
    if (SQLGetData (hstmt, i+1, SQL_CHAR, buffer, sizeof(buffer), &len ) == SQL_SUCCESS )
      return buffer;
  }
  return LString();
}

LString LDBQuery_ODBC::getField ( long i ) const
{
  LString fname = getFieldName ( i );
  if ( internalHash.isSet ( fname ) )
    return internalHash.get ( fname );
  else
    return LString();
}

LString LDBQuery_ODBC::getField ( LString const& name ) const
{
  return getField ( getFieldIdx ( name ) );
}

LString LDBQuery_ODBC::getFieldName ( long i ) const
{
  if ( !isValid() || !hasResult() )
    return LString();
  if ( i > numFields() || i < 0 )
    return LString();

  char buffer [4*1024];

  if ( SQLDescribeCol ( hstmt, i+1, (unsigned char*)buffer, sizeof(buffer), NULL, NULL, NULL, NULL, NULL ) == SQL_SUCCESS );
		return buffer;
  
  return LString();
}

long LDBQuery_ODBC::getFieldIdx ( LString const & name ) const
{
  if ( (!isValid()) || (!hasResult()) ) {
    return -1;
  }

  for ( int i = 0; i<numFields(); i++ ) {
    if ( getFieldName(i).toUpper() == name.toUpper() )
      return i;
  }

  return -1;
}

LString LDBDriver_ODBC::getLastODBCError () const
{
  LTRACE_FUNC ( "LDBDriver_ODBC", "LString getLastODBCError ( )" );
  SWORD len;
  char buffer [4096];
  SQLError ( henv, hdbc, SQL_NULL_HSTMT, NULL, NULL, (unsigned char*)buffer, sizeof(buffer), &len);
  return buffer;
}

LString LDBQuery_ODBC::getLastODBCError () const
{
  LTRACE_FUNC ( "LDBQuery_ODBC", "LString getLastODBCError ( )" );
  SWORD len;
  char buffer [4096];
  SQLError ( driver.henv, driver.hdbc, hstmt, NULL, NULL, (unsigned char*)buffer, sizeof(buffer), &len);
  return buffer;
}

LCL_PLUGIN_INIT("ODBC", "A LDBSystem ODBC PlugIn", 2);
LCL_PLUGIN_FACTORY_WITH_ARGS(LDBDriver_ODBC);

//#endif //HAVE_ODBC
